from __future__ import annotations

from decimal import Decimal

import pandas as pd
from django.apps import apps as django_apps
from django.contrib.sites.models import Site
from django.db.models import DecimalField, ExpressionWrapper, F
from django_pandas.io import read_frame


def in_stock_for_subjects_df(
    container_name: str = None, product_name: str = None
) -> pd.DataFrame:
    """Returns a dataframe of stock allocated to a subject_identifier.

    Filter by site_id to keep those rows already at a study site.

        df_subject_stock = df_subject_stock[
            df_subject_stock.site_id.notna()
        ]


    """
    location_cls = django_apps.get_model("edc_pharmacy.location")
    stock_cls = django_apps.get_model("edc_pharmacy.stock")

    # qs of stock in stock and allocated to a subject_identifier
    difference = ExpressionWrapper(F("qty_in") - F("qty_out"), output_field=DecimalField())
    qs = (
        stock_cls.objects.filter(
            allocation__isnull=False,
            confirmation__isnull=False,
            stocktransferitem__isnull=False,
            confirmationatsiteitem__isnull=False,
            container__name=container_name,
            dispenseitem__isnull=True,
            product__name__icontains=product_name,
        )
        .values(
            "stock_identifier",
            "code",
            "allocation__registered_subject__subject_identifier",
            "qty_in",
            "qty_out",
            "container__name",
            "location",
            "product",
        )
        .annotate(container_qty=difference)
        .filter(qty__gte=Decimal("1.00"))
    )

    # read df
    df = read_frame(qs, verbose=False).rename(
        columns={
            "allocation__registered_subject__subject_identifier": "subject_identifier",
            "container__name": "container",
        }
    )
    df["in_stock"] = True

    # merge in site_id
    df_sites = read_frame(Site.objects.all(), verbose=False)
    df_sites = df_sites.rename(columns={"id": "site_id"})
    df_location = read_frame(location_cls.objects.all(), verbose=False)
    df_location = df_location.merge(df_sites[["name", "site_id"]], on="name", how="left")
    df_location = df_location.rename(columns={"id": "location"})
    df = df.merge(df_location[["location", "site_id"]], how="left", on="location")

    # sort
    df = df.sort_values(by=["subject_identifier", "stock_identifier"])
    df = df.reset_index(drop=True)
    return df



from django.apps import apps as django_apps
from django.db import models

from edc_model.models import BaseUuidModel, HistoricalRecords

MODEL_TYPES = (
    ("consent", "Consent"),
    ("crf", "CRF"),
    ("locator", "Locator"),
    ("plain", "Plain"),
)


class PlanManager(models.Manager):
    def get_by_natural_key(self, name):
        return self.get(name=name)


class Plan(BaseUuidModel):
    name = models.CharField(max_length=25, unique=True)

    model = models.CharField(max_length=50)

    field_names = models.TextField(max_length=500, blank=True, default="")

    header_labels = models.TextField(max_length=500, blank=True, default="")

    lookups = models.TextField(max_length=500, blank=True, default="")

    model_type = models.CharField(max_length=25, choices=MODEL_TYPES, default="plain")

    excluded_field_names = models.TextField(max_length=500, blank=True, default="")

    header = models.BooleanField(verbose_name="Include header", default=True)

    encrypt = models.BooleanField(verbose_name="Mask encrypted values", default=True)

    objects = PlanManager()

    history = HistoricalRecords()

    def __str__(self):
        return f"{self.name} {self.model}"

    def save(self, *args, **kwargs):
        model_cls = django_apps.get_model(self.model)
        if not self.field_names:
            self.field_names = [f.name for f in model_cls._meta.get_fields()]
            self.field_names = ",".join(self.field_names)
        if not self.header_labels:
            self.header_labels = [f.name for f in model_cls._meta.get_fields()]
            self.header_labels = ",".join(self.header_labels)
        super().save(*args, **kwargs)

    def natural_key(self):
        return (self.name,)



import copy
import sys

from django.apps import apps as django_apps
from django.utils.module_loading import import_module, module_has_submodule


class AlreadyRegistered(Exception):
    pass


class RegistryNotLoaded(Exception):
    pass


class SiteDataframeHandlerNotFound(Exception):
    pass


__all__ = ["site_dataframe_handlers"]


class Site:
    def __init__(self):
        self.registry = {}
        self.model_reference_registry = {}
        self.loaded = False

    def register(self, handler_cls=None, app_labels=None, models=None):
        app_labels = app_labels or []
        models = models or []
        name = handler_cls.__name__
        if name in self.registry:
            raise AlreadyRegistered(
                f"Dataframe handler class already been registered. " f"Got {name}"
            )
        self.registry.update({name: handler_cls})
        for app_label in app_labels:
            self.model_reference_registry.update({app_label: name})
        for model in models:
            self.model_reference_registry.update({model: name})
        self.loaded = True

    def get_for(self, key=None):
        if not self.loaded:
            raise RegistryNotLoaded("site_dataframe_handlers not loaded")
        handler_cls = None
        try:
            handler_cls = self.registry[key]
        except KeyError:
            pass
        try:
            handler_cls = self.registry[self.model_reference_registry.get(key)]
        except KeyError:
            pass
        try:
            handler_cls = self.registry[self.model_reference_registry.get(key.split(".")[0])]
        except KeyError:
            pass
        if not handler_cls:
            keys = list(self.registry.keys()) + list(self.model_reference_registry.keys())
            raise SiteDataframeHandlerNotFound(
                f"Cannot find dataframe hander. Got {key}. " f"Expected one of {keys}."
            )
        return handler_cls

    def autodiscover(self, module_name=None):
        """Autodiscovers classes in the site_dataframe_handlers.py file of any
        INSTALLED_APP.
        """
        module_name = module_name or "site_dataframe_handlers"
        sys.stdout.write(f" * checking for {module_name} ...\n")
        for app in django_apps.app_configs:
            try:
                mod = import_module(app)
                try:
                    before_import_registry = copy.copy(site_dataframe_handlers.registry)
                    import_module(f"{app}.{module_name}")
                    sys.stdout.write(
                        f" * registered dataframe handler from application '{app}'\n"
                    )
                except (ImportError, ModuleNotFoundError) as e:
                    if f"No module named '{app}.{module_name}'" not in str(e):
                        site_dataframe_handlers.registry = before_import_registry
                        if module_has_submodule(mod, module_name):
                            raise
            except ModuleNotFoundError:
                pass


site_dataframe_handlers = Site()



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from edc_dashboard.url_names import url_names

from ..filters import ListboardViewFilters

if TYPE_CHECKING:
    from django.db.models import Q


class ListboardFilterViewMixin:
    listboard_view_filters = ListboardViewFilters()
    listboard_filter_url = None

    def __init__(self, **kwargs):
        self.listboard_view_exclude_filter_applied = False  # TODO: ??
        self.listboard_view_include_filter_applied = False  # TODO: ??
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        listboard_url = getattr(self, "listboard_url", None)
        kwargs.update(
            listboard_view_filters=self.listboard_view_filters.filters,
            listboard_filter_url=url_names.get(self.listboard_filter_url or listboard_url),
        )
        return super().get_context_data(**kwargs)

    def get_queryset_filter_options(self, request, *args, **kwargs) -> tuple[Q, dict]:
        q_object, options = super().get_queryset_filter_options(request, *args, **kwargs)
        self.listboard_view_include_filter_applied = False
        for listboard_filter in self.listboard_view_filters.include_filters:
            if self.request.GET.get(listboard_filter.attr) == listboard_filter.name:
                lookup_options = listboard_filter.lookup_options
                if lookup_options:
                    options.update(**listboard_filter.lookup_options)
                self.listboard_view_include_filter_applied = True
        if (
            not self.listboard_view_include_filter_applied
            and self.listboard_view_filters.default_include_filter
        ):
            options.update(**self.listboard_view_filters.default_include_filter.lookup_options)
        return q_object, options

    def get_queryset_exclude_options(self, request, *args, **kwargs) -> tuple[Q, dict]:
        q_object, options = super().get_queryset_exclude_options(request, *args, **kwargs)
        self.listboard_view_exclude_filter_applied = False
        for listboard_filter in self.listboard_view_filters.exclude_filters:
            if self.request.GET.get(listboard_filter.attr) == listboard_filter.name:
                lookup_options = listboard_filter.lookup_options
                if lookup_options:
                    options.update(**listboard_filter.lookup_options)
                self.listboard_view_exclude_filter_applied = True
        if (
            not self.listboard_view_exclude_filter_applied
            and not self.listboard_view_include_filter_applied
            and self.listboard_view_filters.default_exclude_filter
        ):
            options.update(**self.listboard_view_filters.default_exclude_filter.lookup_options)
        return q_object, options



from __future__ import annotations

from typing import TYPE_CHECKING
from urllib import parse
from uuid import UUID

import numpy as np
import pandas as pd
from django.conf import settings
from sqlalchemy import TextClause, create_engine, text

from .dialects import MysqlDialect

if TYPE_CHECKING:
    from sqlalchemy import Engine


class DatabaseNameError(Exception):
    pass


class Database:
    dialect_cls = MysqlDialect
    lowercase_columns = True
    DATABASES_NAME = "default"

    def __init__(self):
        self._database = None
        self._engine: Engine | None = None
        self._tables = pd.DataFrame()
        self.dialect = self.dialect_cls(dbname=self.database)

    @property
    def engine(self) -> Engine:
        if not self._engine:
            db_settings = settings.DATABASES.get("default")
            user = db_settings.get("USER")
            pwd = parse.quote(db_settings.get("PASSWORD"))
            name = db_settings.get("NAME")
            host = db_settings.get("HOST")
            port = db_settings.get("PORT")
            uri = f"mysql://{user}:{pwd}@{host}:{port}/{name}?charset=utf8"
            self._engine = create_engine(uri)
        return self._engine

    @property
    def database(self) -> dict:
        """Returns the database name."""
        return settings.DATABASES[self.DATABASES_NAME]["NAME"]

    def read_sql(
        self, sql: TextClause, params: list | tuple | dict | None = None
    ) -> pd.DataFrame:
        """Returns a dataframe.

        A simple wrapper for pd.read_sql().
        """
        with self.engine.connect() as conn:
            df = pd.read_sql(sql, conn, params=params)
        return df

    def show_databases(self) -> pd.DataFrame:
        """Returns a dataframe of database names in the schema."""
        sql, params = self.dialect.show_databases()
        return self.read_sql(sql, params=params)

    def select_table(
        self,
        table_name: str = None,
        lowercase_columns: list[str] | None = None,
        uuid_columns: list[str] | None = None,
        limit: int | None = None,
    ) -> pd.DataFrame:
        """Returns a dataframe of a table.

        Note: UUID columns are stored as strings (uuid.hex) and need
        to be converted from string to UUID if to match the
        rendering of the same column by a Django model class.
        """
        uuid_columns = uuid_columns or []
        lowercase_columns = lowercase_columns or self.lowercase_columns
        sql, params = self.dialect.select_table(table_name)
        if limit:
            sql = text(f"{sql.text} LIMIT {int(limit)}")
        df = self.read_sql(sql, params=params)
        if lowercase_columns:
            columns = {col: col.lower() for col in list(df.columns)}
            df.rename(columns=columns, inplace=True)
        for col in uuid_columns:
            df[col] = df.apply(lambda row: str(UUID(row[col])) if row[col] else np.nan, axis=1)
        return df

    def show_tables(self, app_label: str = None) -> pd.DataFrame:
        """Returns a dataframe of table names in the schema."""
        sql, params = self.dialect.show_tables(app_label)
        return self.read_sql(sql, params=params)

    def show_tables_with_columns(
        self, app_label: str = None, column_names: list[str] = None
    ) -> pd.DataFrame:
        """Returns a dataframe of table names in the schema
        that have a column in column_names.
        """
        sql, params = self.dialect.show_tables_with_columns(app_label, column_names)
        return self.read_sql(sql, params=params)

    def show_tables_without_columns(
        self, app_label: str = None, column_names: list[str] = None
    ) -> pd.DataFrame:
        """Returns a dataframe of table names in the schema.
        that DO NOT have a column in column_names.
        """
        sql, params = self.dialect.show_tables_without_columns(app_label, column_names)
        return self.read_sql(sql, params=params)

    def show_inline_tables(self, referenced_table_name: str = None):
        sql, params = self.dialect.show_inline_tables(referenced_table_name)
        df = self.read_sql(sql, params=params)
        df = df.rename(
            columns={
                "TABLE_NAME": "table_name",
                "REFERENCED_COLUMN_NAME": "referenced_column_name",
                "REFERENCED_TABLE_NAME": "referenced_table_name",
                "COLUMN_NAME": "column_name",
            }
        )
        return df



from django import forms
from django.conf import settings
from django.urls.base import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe

from edc_constants.constants import GRADE4, GRADE5, YES

from ..form_validators import AeInitialFormValidator
from ..utils import get_ae_model
from .ae_modelform_mixin import AeModelFormMixin


class AeInitialModelFormMixin(AeModelFormMixin):
    form_validator_cls = AeInitialFormValidator

    def clean(self):
        cleaned_data = super().clean()
        self.raise_if_followup_exists()
        self.validate_sae_and_grade()
        return cleaned_data

    def validate_sae_and_grade(self):
        """Raise an exception if grade>=4 and user did not
        indicate that this is an SAE.
        """
        if (
            self.cleaned_data.get("ae_grade") in [GRADE4, GRADE5]
            and self.cleaned_data.get("sae") != YES
        ):
            raise forms.ValidationError({"sae": "Invalid. Grade is >= 4"})

    @property
    def changelist_url(self):
        ae_followup_cls = get_ae_model("aefollowup")
        app_label = ae_followup_cls._meta.app_label
        model_name = ae_followup_cls._meta.object_name.lower()
        return reverse(
            f"{settings.ADVERSE_EVENT_ADMIN_SITE}:{app_label}_{model_name}_changelist"
        )

    def raise_if_followup_exists(self):
        """Raise an exception if the AE followup exists
        and the user is attempting to change this form.
        """
        ae_followup_cls = get_ae_model("aefollowup")
        if ae_followup_cls.objects.filter(ae_initial=self.instance.pk).exists():
            url = f"{self.changelist_url}?q={self.instance.action_identifier}"
            raise forms.ValidationError(
                format_html(  # nosec B703, B308
                    "Unable to save. Follow-up reports exist. Provide updates "
                    "to this report using the "
                    '{} instead. See <A href="{}">AE Follow-ups for {}</A>.',
                    ae_followup_cls._meta.verbose_name,
                    mark_safe(url),  # nosec B703, B308
                    mark_safe(str(self.instance)),  # nosec B703, B308
                )
            )

    class Meta:
        help_texts = {
            "subject_identifier": "(read-only)",
            "action_identifier": "(read-only)",
        }
        widgets = {
            "subject_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
            "action_identifier": forms.TextInput(attrs={"readonly": "readonly"}),
        }



from decimal import Decimal

from django.db import models
from django.utils import timezone
from sequences import get_next_value

from edc_model.models import BaseUuidModel, HistoricalRecords

from ...exceptions import InvalidContainer, ReceiveError, ReceiveItemError
from .container import Container
from .lot import Lot
from .order_item import OrderItem
from .receive import Receive


class Manager(models.Manager):
    use_in_migrations = True


class ReceiveItem(BaseUuidModel):
    receive_item_identifier = models.CharField(
        max_length=36,
        unique=True,
        null=True,
        blank=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    receive_item_datetime = models.DateTimeField(default=timezone.now)

    receive = models.ForeignKey(Receive, on_delete=models.PROTECT, null=True, blank=False)

    container = models.ForeignKey(
        Container,
        on_delete=models.PROTECT,
        null=True,
        blank=False,
        limit_choices_to={"may_receive_as": True},
    )

    order_item = models.ForeignKey(OrderItem, on_delete=models.PROTECT, null=True, blank=False)

    lot = models.ForeignKey(
        Lot,
        verbose_name="Batch",
        on_delete=models.PROTECT,
        null=True,
        blank=False,
    )

    name = models.CharField(
        max_length=200, default="", blank=True, help_text="Leave blank to use default"
    )

    qty = models.DecimalField(
        verbose_name="Quantity", null=True, blank=False, decimal_places=2, max_digits=20
    )

    unit_qty = models.DecimalField(
        verbose_name="Unit quantity",
        null=True,
        blank=True,
        decimal_places=2,
        max_digits=20,
        help_text="Quantity x Container.Quantity, e.g. 10 x Bottle of 128 = 1280",
    )

    reference = models.CharField(max_length=150, default="", blank=True)

    comment = models.TextField(default="", blank=True)

    task_id = models.UUIDField(null=True)

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return (
            f"{self.receive_item_identifier}:"
            f"{self.order_item.product.name}"
            f" | {self.container.name}"
        )

    def save(self, *args, **kwargs):
        if not self.receive_item_identifier:
            self.receive_item_identifier = f"{get_next_value(self._meta.label_lower):06d}"
        if not self.receive:
            raise ReceiveItemError("Receive may not be null.")
        if not self.container:
            raise ReceiveItemError("Container may not be null.")
        if not self.order_item:
            raise ReceiveItemError("OrderItem may not be null.")
        if not self.lot:
            raise ReceiveItemError("Lot may not be null.")
        if self.container.qty > Decimal("1.0"):
            self.unit_qty = self.qty * self.container.qty
        else:
            self.unit_qty = self.qty
        if not self.name:
            self.name = f"{self.order_item.product.name} | {self.container.name}"
        if not self.container.may_receive_as:
            raise InvalidContainer(
                "Invalid container. Container is not configured for receiving. "
                f"Got {self.container}"
            )
        if self.order_item.product.assignment != self.lot.assignment:
            raise ReceiveError("Lot number assignment does not match product assignment!")
        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Receive item"
        verbose_name_plural = "Receive items"



from __future__ import annotations

from django.core.exceptions import ObjectDoesNotExist
from django.forms import ValidationError

from edc_model_form.mixins.report_datetime_modelform_mixin import (
    ReportDatetimeModelFormMixin,
)
from edc_registration import get_registered_subject_model_cls


class BaseModelFormMixinError(Exception):
    pass


__all__ = ["BaseModelFormMixin", "BaseModelFormMixinError"]


class BaseModelFormMixin(ReportDatetimeModelFormMixin):
    """Base modelform mixin for edc forms.

    If this is a CRF, use together with the modelform mixin
    from edc-visit-tracking.
    """

    def clean(self) -> dict:
        cleaned_data = super().clean()
        if not self.report_datetime_field_attr:
            raise BaseModelFormMixinError(
                "Attribute `report_datetime_field_attr` Cannot be None. "
                f"See modelform for {self._meta.model}."
            )
        self.validate_subject_identifier()
        return cleaned_data

    def get_subject_identifier(self):
        """Returns subject identifier.

        Assumes a non-CRF with model field subject_identifier.
        """
        subject_identifier = None
        if "subject_identifier" in self.cleaned_data:
            subject_identifier = self.cleaned_data.get("subject_identifier")
        elif self.instance:
            subject_identifier = self.instance.subject_identifier
        if not subject_identifier:
            raise ValidationError("Invalid. Subject identifier cannot be none.")
        return subject_identifier

    def validate_subject_identifier(self) -> None:
        """Validates subject_identifier exists in RegisteredSubject"""
        try:
            get_registered_subject_model_cls().objects.get(
                subject_identifier=self.get_subject_identifier()
            )
        except ObjectDoesNotExist as e:
            raise ValidationError(
                {
                    "subject_identifier": (
                        "Invalid. Subject is not registered. "
                        f"Got {self.get_subject_identifier()}."
                    )
                }
            ) from e



from django.db import models
from django.utils import timezone
from sequences import get_next_value

from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_sites.model_mixins import SiteModelMixin

from ...exceptions import ConfirmAtSiteError
from .confirmation_at_site import ConfirmationAtSite
from .stock import Stock


class Manager(models.Manager):
    use_in_migrations = True


class ConfirmationAtSiteItem(SiteModelMixin, BaseUuidModel):
    confirmation_at_site = models.ForeignKey(ConfirmationAtSite, on_delete=models.PROTECT)

    transfer_confirmation_item_identifier = models.CharField(
        max_length=36,
        unique=True,
        null=True,
        blank=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    transfer_confirmation_item_datetime = models.DateTimeField(default=timezone.now)

    stock = models.OneToOneField(Stock, on_delete=models.PROTECT)

    confirmed_datetime = models.DateTimeField(null=True, blank=True)

    confirmed_by = models.CharField(max_length=150, default="", blank=True)

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return f"{self.transfer_confirmation_item_identifier} {self.stock.code}"

    def save(self, *args, **kwargs):
        self.site = self.confirmation_at_site.site
        if not self.transfer_confirmation_item_identifier:
            next_id = get_next_value(self._meta.label_lower)
            self.transfer_confirmation_item_identifier = f"{next_id:010d}"
        if (
            self.confirmation_at_site.location.site
            != self.stock.allocation.registered_subject.site
        ):
            raise ConfirmAtSiteError(
                "Location mismatch. Cannot confirm stock item at this location. "
                f"Got {self.stock}."
            )
        super().save(*args, **kwargs)

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Stock Confirmation at Site Item"
        verbose_name_plural = "Stock Confirmation at Site Items"



from django.db import models
from django.utils import timezone

from edc_constants.choices import YES_NO, YES_NO_NA
from edc_constants.constants import NOT_APPLICABLE
from edc_model.validators import datetime_not_future
from edc_model_fields.fields import OtherCharField

from ...models import AeClassification
from ...utils import get_adverse_event_app_label


def get_investigator_ae_classification_choices():
    choices = [(obj.name, obj.display_name) for obj in AeClassification.objects.all()]
    choices.append((NOT_APPLICABLE, "Not applicable"))
    return tuple(choices)


class AeTmgFieldsModelMixin(models.Model):
    ae_initial = models.ForeignKey(
        f"{get_adverse_event_app_label()}.aeinitial", on_delete=models.PROTECT
    )

    report_datetime = models.DateTimeField(
        verbose_name="Report date and time",
        validators=[datetime_not_future],
        default=timezone.now,
    )

    ae_received_datetime = models.DateTimeField(
        blank=True,
        null=True,
        validators=[datetime_not_future],
        verbose_name="Date and time AE form received:",
    )

    clinical_review_datetime = models.DateTimeField(
        blank=True,
        null=True,
        validators=[datetime_not_future],
        verbose_name="Date and time of clinical review: ",
    )

    ae_description = models.TextField(
        verbose_name="Description of AE:",
        default="",
        blank=True,
    )

    investigator_comments = models.TextField(
        verbose_name="This investigator's comments:",
        default="",
        blank=True,
    )

    ae_classification = models.CharField(max_length=150, blank=True, default="")

    ae_classification_other = OtherCharField(max_length=250, blank=True, default="")

    original_report_agreed = models.CharField(
        verbose_name="Does this investigator agree with the original AE report?",
        max_length=15,
        choices=YES_NO,
        default="",
        help_text="If No, explain in the narrative below",
    )

    investigator_narrative = models.TextField(verbose_name="Narrative", blank=True, default="")

    investigator_ae_classification_agreed = models.CharField(
        verbose_name=(
            "Does this investigator agree with the AE classification on the "
            "original AE report?"
        ),
        max_length=15,
        choices=YES_NO_NA,
        default=NOT_APPLICABLE,
        help_text="If No, select a classification below",
    )

    investigator_ae_classification = models.ForeignKey(
        AeClassification,
        on_delete=models.PROTECT,
        verbose_name="Adverse Event (AE) Classification",
        null=True,
        help_text=(
            "Only applicable if this investigator does not agree with the original AE report"
        ),
    )

    investigator_ae_classification_other = OtherCharField(
        max_length=250, blank=True, default=""
    )

    officials_notified = models.DateTimeField(
        blank=True,
        null=True,
        validators=[datetime_not_future],
        verbose_name="Date and time regulatory authorities notified (SUSARs)",
    )

    class Meta:
        abstract = True



from reportlab.lib import colors
from reportlab.platypus import Flowable


class CheckboxFlowable(Flowable):
    def __init__(self, name, size=10, checked=False):
        Flowable.__init__(self)
        self.name = name
        self.size = size
        self.checked = checked

    def draw(self):
        x = self.canv._currentMatrix[4] + self.size / 2 - 10
        y = self.canv._currentMatrix[5] - self.size / 2
        self.canv.acroForm.checkbox(
            name=self.name,
            tooltip=self.name,
            size=self.size,
            x=x,
            y=y,
            buttonStyle="check",
            borderStyle="solid",
            borderWidth=0.5,
            borderColor=colors.black,
            fillColor=colors.white,
            textColor=colors.black,
            forceBorder=True,
            checked=self.checked,
        )



# Generated by Django 2.2.2 on 2019-07-06 05:06

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_metadata", "0012_auto_20190627_2320")]

    operations = [
        migrations.RemoveIndex(
            model_name="crfmetadata", name="edc_metadat_subject_a76082_idx"
        ),
        migrations.RemoveIndex(
            model_name="requisitionmetadata", name="edc_metadat_subject_f96783_idx"
        ),
        migrations.AddField(
            model_name="crfmetadata",
            name="timepoint",
            field=models.DecimalField(decimal_places=1, max_digits=6, null=True),
        ),
        migrations.AddField(
            model_name="requisitionmetadata",
            name="timepoint",
            field=models.DecimalField(decimal_places=1, max_digits=6, null=True),
        ),
        migrations.AddIndex(
            model_name="crfmetadata",
            index=models.Index(
                fields=[
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "visit_code_sequence",
                    "timepoint",
                    "model",
                    "entry_status",
                    "show_order",
                ],
                name="edc_metadat_subject_1b2c7e_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="requisitionmetadata",
            index=models.Index(
                fields=[
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "visit_code_sequence",
                    "timepoint",
                    "model",
                    "entry_status",
                    "show_order",
                    "panel_name",
                ],
                name="edc_metadat_subject_fb1a00_idx",
            ),
        ),
    ]



# Generated by Django 6.0 on 2025-03-13 11:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_pharmacy",
            "0063_alter_allocation_managers_remove_allocation_site_and_more",
        ),
    ]

    operations = [
        migrations.AddField(
            model_name="historicalstockrequest",
            name="start_datetime",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="If provided, will not include appointments before this date. May be left blank",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="stockrequest",
            name="start_datetime",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="If provided, will not include appointments before this date. May be left blank",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="historicalstockrequest",
            name="cutoff_datetime",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="If provided, will not include appointments after this date. May be left blank",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="stockrequest",
            name="cutoff_datetime",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="If provided, will not include appointments after this date. May be left blank",
                null=True,
            ),
        ),
    ]



# Generated by Django 3.2.13 on 2022-09-27 00:49

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_protocol_incident", "0012_auto_20220913_2139"),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name="actionsrequired",
            name="edc_protoco_id_158eab_idx",
        ),
        migrations.RemoveIndex(
            model_name="protocoldeviationviolation",
            name="edc_protoco_subject_e56b31_idx",
        ),
        migrations.RemoveIndex(
            model_name="protocolincident",
            name="edc_protoco_subject_6848b1_idx",
        ),
        migrations.RemoveIndex(
            model_name="protocolincidents",
            name="edc_protoco_id_7ee7c7_idx",
        ),
        migrations.RemoveIndex(
            model_name="protocolviolations",
            name="edc_protoco_id_fe21ea_idx",
        ),
        migrations.AddIndex(
            model_name="actionsrequired",
            index=models.Index(
                fields=["id", "display_name", "display_index"],
                name="edc_protoco_id_17a455_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="protocoldeviationviolation",
            index=models.Index(
                fields=["subject_identifier", "action_identifier", "site", "id"],
                name="edc_protoco_subject_366adb_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="protocolincident",
            index=models.Index(
                fields=["subject_identifier", "action_identifier", "site", "id"],
                name="edc_protoco_subject_1c6fcb_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="protocolincidents",
            index=models.Index(
                fields=["id", "display_name", "display_index"],
                name="edc_protoco_id_7c6a33_idx",
            ),
        ),
        migrations.AddIndex(
            model_name="protocolviolations",
            index=models.Index(
                fields=["id", "display_name", "display_index"],
                name="edc_protoco_id_58f1f5_idx",
            ),
        ),
    ]



from ..constants import UPDATE
from .model_notification import ModelNotification


class UpdatedModelNotification(ModelNotification):
    model_operations = (UPDATE,)

    update_fields = ("modified",)

    email_subject_template = (
        "*UPDATE* {test_subject_line}{protocol_name}: "
        "{display_name} "
        "for {instance.subject_identifier}"
    )



# -*- coding: utf-8 -*-
# Generated by Django 1.10.3 on 2016-12-21 21:35
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_appointment", "0004_auto_20161210_1408")]

    operations = [
        migrations.AlterField(
            model_name="appointment",
            name="subject_identifier",
            field=models.CharField(max_length=50, verbose_name="Subject Identifier"),
        ),
        migrations.AlterField(
            model_name="historicalappointment",
            name="subject_identifier",
            field=models.CharField(max_length=50, verbose_name="Subject Identifier"),
        ),
    ]



from django.db import models
from django.db.models.deletion import PROTECT

from edc_model.models import BaseUuidModel, HistoricalRecords
from edc_pdf_reports.model_mixins import PdfReportModelMixin
from edc_search.model_mixins import SearchSlugManager, SearchSlugModelMixin
from edc_sites.managers import CurrentSiteManager

from ...managers import ManifestManager
from ...model_mixins import ManifestModelMixin
from ...pdf_reports import ManifestPdfReport
from .consignee import Consignee
from .shipper import Shipper


class Manager(ManifestManager, SearchSlugManager):
    pass


class Manifest(ManifestModelMixin, PdfReportModelMixin, SearchSlugModelMixin, BaseUuidModel):
    pdf_report_cls = ManifestPdfReport

    def get_search_slug_fields(self):
        return (
            "manifest_identifier",
            "human_readable_identifier",
            "shipper.name",
            "consignee.name",
        )

    consignee = models.ForeignKey(Consignee, verbose_name="Consignee", on_delete=PROTECT)

    shipper = models.ForeignKey(Shipper, verbose_name="Shipper/Exporter", on_delete=PROTECT)

    objects = Manager()

    on_site = CurrentSiteManager()

    history = HistoricalRecords()

    def natural_key(self):
        return (self.manifest_identifier,)

    natural_key.dependencies = ("edc_lab.shipper", "edc_lab.consignee")

    def __str__(self):
        return "{} created on {} by {}".format(
            self.manifest_identifier,
            self.manifest_datetime.strftime("%Y-%m-%d"),
            self.user_created,
        )

    @property
    def count(self):
        return self.manifestitem_set.all().count()

    class Meta(ManifestModelMixin.Meta, BaseUuidModel.Meta):
        verbose_name = "Manifest"



from django.contrib import admin
from django_audit_fields import audit_fieldset_tuple

from ...admin_site import edc_pharmacy_admin
from ...auth_objects import PHARMACIST_ROLE, PHARMACY_SUPER_ROLE
from ...models import StockProxy
from ..list_filters import (
    ConfirmedAtSiteFilter,
    DispensedFilter,
    StoredAtSiteFilter,
    TransferredListFilter,
)
from .stock_admin import StockAdmin


@admin.register(StockProxy, site=edc_pharmacy_admin)
class StockProxyAdmin(StockAdmin):

    fieldsets = (
        (
            "Stock item",
            {
                "fields": (
                    "stock_identifier",
                    "code",
                    "location",
                )
            },
        ),
        (
            "Quantity",
            {"fields": ("qty_in", "qty_out", "unit_qty_in", "unit_qty_out")},
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "formatted_code",
        "formatted_transferred",
        "formatted_confirmed_at_site",
        "formatted_stored_at_site",
        "dispensed",
        "stock_request_changelist",
        "stock_transfer_item_changelist",
        "allocation_changelist",
        "dispense_changelist",
        "formulation",
        "qty",
        "container_str",
        "unit_qty",
        "created",
        "modified",
    )
    list_filter = (
        TransferredListFilter,
        ConfirmedAtSiteFilter,
        StoredAtSiteFilter,
        DispensedFilter,
        "product__formulation__description",
        "location__display_name",
        "created",
        "modified",
    )
    search_fields = (
        "stock_identifier",
        "from_stock__stock_identifier",
        "code",
        "from_stock__code",
        "repack_request__id",
        "allocation__registered_subject__subject_identifier",
        "allocation__stock_request_item__id",
        "allocation__stock_request_item__stock_request__id",
        "allocation__id",
        "stocktransferitem__stock_transfer__id",
    )
    readonly_fields = (
        "code",
        "confirmation",
        "container",
        "from_stock",
        "location",
        "repack_request",
        "lot",
        "product",
        "qty_in",
        "qty_out",
        "unit_qty_in",
        "unit_qty_out",
        "receive_item",
        "stock_identifier",
    )

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .filter(
                confirmation__isnull=False,
                allocation__isnull=False,
                container__may_request_as=True,
            )
        )

    def get_list_display_links(self, request, list_display):
        display_links = super().get_list_display_links(request, list_display)
        if not request.user.userprofile.roles.filter(
            name__in=[PHARMACIST_ROLE, PHARMACY_SUPER_ROLE]
        ).exists():
            try:
                display_links.remove("formatted_code")
            except ValueError:
                pass
        return display_links

    def get_view_only_site_ids_for_user(self, request) -> list[int]:
        return [s.id for s in request.user.userprofile.sites.all() if s.id != request.site.id]



# Generated by Django 5.1.2 on 2024-11-18 02:17

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0040_historicalstock_dispensed_stock_dispensed"),
    ]

    operations = [
        migrations.AlterField(
            model_name="dispenseitem",
            name="dispense",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="edc_pharmacy.dispense",
                verbose_name="Dispense",
            ),
        ),
        migrations.AlterField(
            model_name="historicaldispenseitem",
            name="dispense",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="edc_pharmacy.dispense",
                verbose_name="Dispense",
            ),
        ),
    ]



from django.db import models
from django.db.models import UniqueConstraint

from edc_model.models import BaseUuidModel
from edc_sites.managers import CurrentSiteManager

from .. import KEYED, NOT_REQUIRED, REQUIRED
from ..managers import RequisitionMetadataManager
from .crf_metadata_model_mixin import CrfMetadataModelMixin


class RequisitionMetadata(CrfMetadataModelMixin, BaseUuidModel):
    panel_name = models.CharField(max_length=50, default="")

    objects = RequisitionMetadataManager()

    on_site = CurrentSiteManager()

    def __str__(self) -> str:
        return (
            f"RequisitionMeta {self.model} {self.visit_schedule_name}."
            f"{self.schedule_name}.{self.visit_code}.{self.visit_code_sequence}@"
            f"{self.timepoint} {self.panel_name} {self.entry_status} "
            f"{self.subject_identifier}"
        )

    @property
    def verbose_name(self) -> str:
        from edc_lab.site_labs import site_labs  # noqa: PLC0415

        return site_labs.panel_names.get(self.panel_name) or self.panel_name

    def natural_key(self) -> tuple:
        return (
            self.panel_name,
            self.model,
            self.subject_identifier,
            self.visit_schedule_name,
            self.schedule_name,
            self.visit_code,
            self.visit_code_sequence,
        )

    # noinspection PyTypeHints
    natural_key.dependencies = ("sites.Site",)  # type: ignore

    def model_instance_query_opts(self) -> dict:
        opts = super().model_instance_query_opts()
        opts.update(panel__name=self.panel_name)
        return opts

    def refresh_entry_status(self) -> str:
        """Resets entry_status to the original visit schedule value"""
        if not self.model_instance:
            if self.visit_code_sequence > 0:
                self.entry_status = REQUIRED
            else:
                visit = self.visits.get(self.visit_code)
                for required in [
                    req.required
                    for req in visit.requisitions
                    if req.model == self.model and req.panel.name == self.panel_name
                ]:
                    self.entry_status = REQUIRED if required else NOT_REQUIRED
        else:
            self.entry_status = KEYED
        self.save_base(update_fields=["entry_status"])
        return self.entry_status

    class Meta(CrfMetadataModelMixin.Meta, BaseUuidModel.Meta):
        app_label = "edc_metadata"
        verbose_name = "Requisition collection status"
        verbose_name_plural = "Requisition collection status"
        constraints = (
            UniqueConstraint(
                fields=[
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "visit_code_sequence",
                    "model",
                    "panel_name",
                ],
                name="%(app_label)s_%(class)s_subject_iden_visit_uniq",
            ),
        )
        indexes = (
            *CrfMetadataModelMixin.Meta.indexes,
            *BaseUuidModel.Meta.indexes,
            models.Index(
                fields=[
                    "site",
                    "entry_status",
                    "visit_code",
                    "visit_code_sequence",
                    "model",
                    "panel_name",
                    "subject_identifier",
                    "schedule_name",
                    "visit_schedule_name",
                ],
            ),
            models.Index(
                fields=[
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "visit_code_sequence",
                    "model",
                    "panel_name",
                    "entry_status",
                    "timepoint",
                    "show_order",
                ],
            ),
        )



# Generated by Django 4.2.11 on 2024-06-17 22:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_qareports", "0002_rename_reportnote_qareportnote_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="qareportnote",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "QA Report Note",
                "verbose_name_plural": "QA Report Notes",
            },
        ),
        migrations.RemoveConstraint(
            model_name="qareportnote",
            name="edc_qareports_qareportnote_name_subj_uniq",
        ),
        migrations.AddConstraint(
            model_name="qareportnote",
            constraint=models.UniqueConstraint(
                fields=("report_model", "subject_identifier"),
                name="edc_qareports_qareportnote_report_model_subj_uniq",
            ),
        ),
    ]



from django.db import models


class AeInitialTmgModelMixin(models.Model):
    tmg_report_datetime = models.DateTimeField(
        verbose_name="Date and time AE reported to TMG",
        blank=True,
        null=True,
        help_text=(
            "AEs ≥ Grade 4 or SAE must be reported to the Trial "
            "Management Group (TMG) within 24 hours"
        ),
    )

    class Meta:
        abstract = True



from django.core.management.base import BaseCommand

from edc_crf.update_crf_status_command import update_crf_status_command


class Command(BaseCommand):
    def handle(self, *args, **options):
        update_crf_status_command(app_label=None)



from django.db import models


class RegisteredSubjectManager(models.Manager):
    use_in_migrations = True

    def get_by_natural_key(self, subject_identifier_as_pk):
        return self.get(subject_identifier_as_pk=subject_identifier_as_pk)

    def get_for_subject_identifier(self, subject_identifier):
        """Returns a queryset for the given subject_identifier."""
        options = {"subject_identifier": subject_identifier}
        return self.filter(**options)

    def get_for_visit(self, visit):
        options = {"subject_identifier": visit.subject_identifier}
        return self.get(**options)



from django.contrib.messages import get_messages


def message_in_queue(request, message_text):
    storage = get_messages(request)
    return any(message.message == message_text for message in storage)



from django.apps.config import AppConfig as DjangoAppConfig
from django.core.management.color import color_style

style = color_style()


class AppConfig(DjangoAppConfig):
    name = "edc_metadata"
    verbose_name = "Data Collection Status"
    metadata_rules_enabled = True
    has_exportable_data = True
    include_in_administration_section = True



# -*- coding: utf-8 -*-
# Generated by Django 1.10.4 on 2016-12-10 14:08
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("edc_appointment", "0003_auto_20161127_2226")]

    operations = [
        migrations.AlterModelOptions(
            name="appointment", options={"ordering": ("timepoint_datetime",)}
        ),
        migrations.AlterUniqueTogether(
            name="appointment",
            unique_together=set(
                [
                    (
                        "subject_identifier",
                        "visit_schedule_name",
                        "schedule_name",
                        "visit_code",
                        "timepoint",
                    ),
                    (
                        "subject_identifier",
                        "visit_schedule_name",
                        "schedule_name",
                        "visit_code",
                        "visit_code_sequence",
                    ),
                ]
            ),
        ),
    ]



from decimal import Decimal

from django.contrib import admin, messages
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _
from django_audit_fields.admin import audit_fieldset_tuple

from edc_constants.constants import NEW, PARTIAL, RECEIVED
from edc_model_admin.history import SimpleHistoryAdmin
from edc_utils import message_in_queue

from ...admin_site import edc_pharmacy_admin
from ...forms import OrderItemForm
from ...models import Order, OrderItem, Receive, Stock
from ...utils import format_qty
from ..actions import delete_order_items_action
from ..list_filters import OrderItemStatusListFilter, ProductAssignmentListFilter
from ..model_admin_mixin import ModelAdminMixin
from ..remove_fields_for_blinded_users import remove_fields_for_blinded_users


@admin.register(OrderItem, site=edc_pharmacy_admin)
class OrderItemAdmin(ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Ordered items"
    history_list_display = ()
    show_object_tools = False
    show_cancel = True
    list_per_page = 20

    form = OrderItemForm
    ordering = ("-order_item_identifier",)
    autocomplete_fields = ["product", "container"]
    actions = [delete_order_items_action]

    fieldsets = (
        (
            None,
            {"fields": (["order", "product", "container"])},
        ),
        (
            "Quantity",
            {"fields": (["qty", "unit_qty", "unit_qty_received"])},
        ),
        ("Status", {"fields": (["status"])}),
        audit_fieldset_tuple,
    )

    list_display = (
        "identifier",
        "product__name",
        "assignment",
        "container",
        "formatted_qty",
        "formatted_unit_qty",
        "order_status",
        "order_changelist",
        "receive_changelist",
        "stock_changelist",
        "created",
        "modified",
    )
    list_filter = (
        OrderItemStatusListFilter,
        ProductAssignmentListFilter,
    )
    radio_fields = {"status": admin.VERTICAL}
    search_fields = (
        "id",
        "order__id",
        "order__order_identifier",
    )
    readonly_fields = (
        "unit_qty",
        "unit_qty_received",
    )

    def get_list_display(self, request):
        fields = super().get_list_display(request)
        fields = remove_fields_for_blinded_users(request, fields)
        return fields

    def get_list_filter(self, request):
        fields = super().get_list_filter(request)
        fields = remove_fields_for_blinded_users(request, fields)
        return fields

    def get_search_fields(self, request):
        fields = super().get_search_fields(request)
        fields = remove_fields_for_blinded_users(request, fields)
        return fields

    def get_readonly_fields(self, request, obj=None):
        if obj:
            return self.readonly_fields + ("order",)
        return self.readonly_fields

    def get_queryset(self, request):
        queryset = super().get_queryset(request)
        msg = _("All order items have been received")
        if not message_in_queue(request, msg):
            if (
                queryset.values("unit_qty").filter(unit_qty=0).count()
                == queryset.model.objects.values("unit_qty").all().count()
            ):
                messages.add_message(request, messages.INFO, msg)
        return queryset

    @admin.display(description="ORDER ITEM #", ordering="-order_item_identifier")
    def identifier(self, obj):
        return obj.order_item_identifier

    @admin.display(description="Assignment", ordering="product__assignment__name")
    def assignment(self, obj):
        return obj.product.assignment

    @admin.display(description="QTY", ordering="qty")
    def formatted_qty(self, obj):
        return format_qty(obj.qty, obj.container)

    @admin.display(description="Units", ordering="qty")
    def formatted_unit_qty(self, obj):
        return format_qty(obj.unit_qty, obj.container)

    @admin.display(description="Product", ordering="product__name")
    def product_name(self, obj):
        return obj.product.formulation.get_description_with_assignment(obj.product.assignment)

    @admin.display(description="Container", ordering="container__name")
    def container_name(self, obj):
        return obj.container.name

    @admin.display(description="Order #", ordering="-order__order_datetime")
    def order_changelist(self, obj):
        url = reverse("edc_pharmacy_admin:edc_pharmacy_order_changelist")
        url = f"{url}?q={obj.order.order_identifier}"
        context = dict(url=url, label=obj.order.order_identifier, title="Back to order")
        return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)

    @admin.display(description="Stock")
    def stock_changelist(self, obj):
        if Stock.objects.filter(receive_item__receive__order=obj.order).exists():
            url = reverse("edc_pharmacy_admin:edc_pharmacy_stock_changelist")
            url = f"{url}?q={obj.order.id}"
            context = dict(url=url, label="Stock", title="Go to stock")
            return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)
        return None

    @admin.display(description="RECEIVE #")
    def receive_changelist(self, obj):
        if receive := self.get_receive_obj(obj):
            return self.render_receive_changelist_link(receive)
        return None

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == "order":
            if request.GET.get("order"):
                kwargs["queryset"] = Order.objects.filter(
                    id__exact=request.GET.get("order", 0)
                )
            else:
                kwargs["queryset"] = Order.objects.none()
        return super().formfield_for_foreignkey(db_field, request, **kwargs)

    def get_status_label(self, obj: OrderItem) -> tuple[str, str]:
        if obj.unit_qty == 0:
            return RECEIVED, _("Received")
        if obj.unit_qty == obj.unit_qty_ordered:
            return NEW, _("New")
        return PARTIAL, _("Partial")

    @admin.display(description="Status")
    def order_status(self, obj: OrderItem):
        receive_this_item_button = ""
        receive = self.get_receive_obj(obj)
        if not receive:
            return self.render_start_receiving_button(obj)
        if (obj.unit_qty_received or Decimal(0)) < obj.unit_qty_ordered:
            receive_this_item_button = self.render_receive_this_item_button(obj, receive)
        received_items_link = self.render_received_items_link(obj)
        renders = [
            receive_this_item_button,
            received_items_link,
        ]
        renders = [r for r in renders if r]
        return format_html(
            "{}",
            mark_safe("<BR>".join(renders)),  # nosec B703, B308
        )

    @staticmethod
    def get_receive_obj(obj: OrderItem) -> Receive | None:
        try:
            obj = Receive.objects.get(order=obj.order)
        except Receive.DoesNotExist:
            obj = None
        return obj

    @staticmethod
    def render_start_receiving_button(obj: OrderItem) -> str:
        url = reverse("edc_pharmacy_admin:edc_pharmacy_receive_add")
        next_url = "edc_pharmacy_admin:edc_pharmacy_orderitem_changelist"
        url = f"{url}?next={next_url}&q={obj.order.id!s}&order={obj.order.id!s}"
        context = dict(
            url=url,
            label=_("Start Receiving"),
            title=_("Receive against this order item"),
        )
        return render_to_string(
            "edc_pharmacy/stock/items_as_button.html",
            context=context,
        )

    @staticmethod
    def render_receive_changelist_link(receive: Receive) -> str:
        url = reverse("edc_pharmacy_admin:edc_pharmacy_receive_changelist")
        url = f"{url}?q={receive.receive_identifier!s}"
        context = dict(url=url, label=receive.receive_identifier, title=_("Receive"))
        return render_to_string(
            "edc_pharmacy/stock/items_as_link.html",
            context=context,
        )

    @staticmethod
    def render_receive_this_item_button(obj: OrderItem, receive: Receive) -> str:
        url = reverse("edc_pharmacy_admin:edc_pharmacy_receiveitem_add")
        next_url = "edc_pharmacy_admin:edc_pharmacy_orderitem_changelist"
        url = (
            f"{url}?next={next_url}&order_item={obj.id!s}&q={obj.order.id!s}"
            f"&receive={receive.id!s}&container={obj.container.id!s}"
        )
        context = dict(
            url=url,
            label=_("Receive this item"),
            title=_("Receive against this order item"),
        )
        return render_to_string(
            "edc_pharmacy/stock/items_as_button.html",
            context=context,
        )

    def render_received_items_link(self, obj: OrderItem) -> str | None:
        status, status_label = self.get_status_label(obj)
        if status != NEW:
            url = reverse("edc_pharmacy_admin:edc_pharmacy_receiveitem_changelist")
            url = f"{url}?q={obj.pk!s}"
            context = dict(url=url, label=status_label, title=_("Go to received items"))
            return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)
        return None



from __future__ import annotations

from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.core.handlers.wsgi import WSGIRequest
from django.db import transaction
from django.utils import timezone

if TYPE_CHECKING:
    from ..models import StockTransfer


def transfer_stock(
    stock_transfer: StockTransfer, stock_codes: list[str], request: WSGIRequest = None
) -> tuple[list[str], list[str], list[str]]:
    stock_model_cls = django_apps.get_model("edc_pharmacy.stock")
    stock_transfer_item_model_cls = django_apps.get_model("edc_pharmacy.stocktransferitem")
    transferred, skipped_codes, invalid_codes = [], [], []
    for stock_code in stock_codes:
        try:
            stock = stock_model_cls.objects.get(
                code=stock_code,
                confirmation__isnull=False,
                allocation__registered_subject__site=stock_transfer.to_location.site,
                location=stock_transfer.from_location,
            )
        except ObjectDoesNotExist:
            skipped_codes.append(stock_code)
        else:
            with transaction.atomic():
                stock_transfer_item_model_cls.objects.create(
                    stock=stock,
                    stock_transfer=stock_transfer,
                    user_created=request.user.username,
                    created=timezone.now(),
                )
                stock.location = stock_transfer.to_location
                stock.save()
                transferred.append(stock_code)
    return transferred, skipped_codes, invalid_codes


__all__ = ["transfer_stock"]



from edc_auth.site_auths import site_auths
from edc_auth.utils import remove_default_model_permissions_from_edc_permissions
from edc_pylabels.auth_objects import PYLABELS

from .auth_objects import (
    PHARMACIST_ROLE,
    PHARMACY,
    PHARMACY_AUDITOR_ROLE,
    PHARMACY_PRESCRIBER,
    PHARMACY_PRESCRIBER_ROLE,
    PHARMACY_SITE,
    PHARMACY_SUPER_ROLE,
    PHARMACY_VIEW,
    SITE_PHARMACIST_ROLE,
    navbar_tuples,
    pharmacy_codenames,
    pharmacy_site_codenames,
    prescriber_codenames,
)


def update_site_auths() -> None:
    site_auths.add_post_update_func(
        "edc_pharmacy", remove_default_model_permissions_from_edc_permissions
    )
    site_auths.add_custom_permissions_tuples(
        model="edc_pharmacy.edcpermissions", codename_tuples=navbar_tuples
    )

    site_auths.add_group(*pharmacy_codenames, name=PHARMACY_VIEW, view_only=True)
    site_auths.add_group(*prescriber_codenames, name=PHARMACY_PRESCRIBER, no_delete=True)
    site_auths.add_group(*pharmacy_site_codenames, name=PHARMACY_SITE, no_delete=False)
    site_auths.add_group(*pharmacy_codenames, name=PHARMACY, no_delete=False)

    site_auths.add_role(PHARMACY_VIEW, name=PHARMACY_AUDITOR_ROLE)
    site_auths.add_role(PHARMACY_PRESCRIBER, name=PHARMACY_PRESCRIBER_ROLE)
    site_auths.add_role(PHARMACY_SITE, PYLABELS, name=SITE_PHARMACIST_ROLE)
    site_auths.add_role(PHARMACY, PYLABELS, name=PHARMACIST_ROLE)
    site_auths.add_role(PHARMACY, name=PHARMACY_SUPER_ROLE)


update_site_auths()



from ...filters import ListboardFilter
from ...filters import ListboardViewFilters as Base


class ScreeningListboardViewFilters(Base):
    all = ListboardFilter(name="all", label="All", lookup={})

    eligible = ListboardFilter(label="Eligible", position=10, lookup={"eligible": True})

    not_eligible = ListboardFilter(
        label="Not Eligible", position=11, lookup={"eligible": False}
    )

    consented = ListboardFilter(
        label="Consented", position=20, lookup={"eligible": True, "consented": True}
    )

    not_consented = ListboardFilter(
        label="Not consented",
        position=21,
        lookup={"eligible": True, "consented": False},
    )



# -*- coding: utf-8 -*-
# Generated by Django 1.10.4 on 2016-12-21 23:23
from __future__ import unicode_literals
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_identifier", "0008_identifiermodel_linked_identifier")]

    operations = [
        migrations.AlterField(
            model_name="historicalidentifierhistory",
            name="created_datetime",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name="identifierhistory",
            name="created_datetime",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
    ]



from .convert_dates_from_model import convert_dates_from_model
from .convert_numbers_to_float64_int64 import convert_numbers_to_float64_int64
from .convert_numbers_to_nullable_dtype import convert_numbers_to_nullable_dtype
from .convert_numerics_from_model import convert_numerics_from_model
from .convert_timedelta_from_model import convert_timedelta_from_model
from .datetime_to_date import datetime_to_date
from .decrypt import DecryptError, decrypt
from .get_export_folder import get_export_folder
from .identity256 import identity256, identity256_decrypt
from .missing_subject_identifiers import missing_subject_identifiers
from .model_from_table_name import get_model_from_table_name
from .refresh_model_from_dataframe import refresh_model_from_dataframe
from .table_names import get_model_names, get_table_names
from .undash import undash



from django.contrib import admin
from django.template.loader import render_to_string
from django.urls import reverse
from django_audit_fields.admin import audit_fieldset_tuple
from rangefilter.filters import DateRangeFilterBuilder

from edc_model_admin.history import SimpleHistoryAdmin

from ...admin_site import edc_pharmacy_admin
from ...models import StockAdjustment
from ..model_admin_mixin import ModelAdminMixin


@admin.register(StockAdjustment, site=edc_pharmacy_admin)
class StockAdjustmentAdmin(ModelAdminMixin, SimpleHistoryAdmin):
    change_list_title = "Pharmacy: Stock adjustment"
    change_form_title = "Pharmacy: Stock adjustment"
    history_list_display = ()
    show_object_tools = True
    show_cancel = True
    list_per_page = 20

    show_form_tools = True
    show_history_label = True
    autocomplete_fields = ["stock"]

    # form = StockAdjustmentForm

    ordering = ("-adjustment_datetime",)

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "stock",
                    "adjustment_datetime",
                    "unit_qty_in_old",
                    "unit_qty_in_new",
                    "reason",
                )
            },
        ),
        audit_fieldset_tuple,
    )

    list_display = (
        "formatted_code",
        "adjustment_datetime",
        "unit_qty_in_old",
        "unit_qty_in_new",
        "reason",
        "created",
        "modified",
    )
    list_filter = (
        ("adjustment_datetime", DateRangeFilterBuilder()),
        "created",
        "modified",
    )
    search_fields = (
        "stock__stock_identifier",
        "stock__code",
    )

    @admin.display(description="Stock #", ordering="stock__code")
    def formatted_code(self, obj):
        return obj.stock.code

    @admin.display(description="From stock #", ordering="from_stock__code")
    def stock_changelist(self, obj):
        if obj.stock:
            url = reverse("edc_pharmacy_admin:edc_pharmacy_stock_changelist")
            url = f"{url}?q={obj.stock.code}"
            context = dict(
                url=url,
                label=obj.stock.code,
                title="Go to stock",
            )
            return render_to_string("edc_pharmacy/stock/items_as_link.html", context=context)
        return None



from __future__ import annotations

from typing import Any

from edc_registration import get_registered_subject
from edc_registration.utils import valid_subject_identifier_or_raise


class RegisteredSubjectViewMixin:
    """Adds the subject_identifier to the context."""

    def __init__(self, **kwargs):
        self._subject_identifier: str | None = None
        self._registered_subject: str | None = None
        super().__init__(**kwargs)

    def get(self, request, *args, **kwargs):
        # subject_identifier will only have value if passed in the url
        self._subject_identifier = kwargs.get("subject_identifier")
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        if self.subject_identifier and valid_subject_identifier_or_raise(
            self.subject_identifier, raise_exception=True
        ):
            kwargs.update(
                subject_identifier=self.registered_subject.subject_identifier,
                gender=self.registered_subject.gender,
                dob=self.registered_subject.dob,
                initials=self.registered_subject.initials,
                identity=self.registered_subject.identity,
                firstname=self.registered_subject.first_name,
                lastname=self.registered_subject.last_name,
                registered_subject=self.registered_subject,
                registered_subject_pk=str(self.registered_subject.id),
            )
        return super().get_context_data(**kwargs)

    @property
    def subject_identifier(self):
        return self._subject_identifier

    @property
    def registered_subject(self):
        if not self._registered_subject:
            self._registered_subject = get_registered_subject(
                self.subject_identifier, raise_exception=True
            )
        return self._registered_subject



# Generated by Django 2.0.1 on 2018-01-16 12:11

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [("edc_identifier", "0015_auto_20180116_1235")]

    operations = [
        migrations.AlterField(
            model_name="identifiermodel",
            name="site",
            field=models.ForeignKey(
                editable=False,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="sites.Site",
            ),
        )
    ]



import warnings
from typing import Any

from django.apps import apps as django_apps
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.utils.translation import gettext as _
from django_revision.views import RevisionMixin

from edc_sites.view_mixins import SiteViewMixin

from .message_view_mixin import MessageViewMixin
from .template_request_context_mixin import TemplateRequestContextMixin


class EdcViewMixin(
    LoginRequiredMixin,
    MessageViewMixin,
    RevisionMixin,
    SiteViewMixin,
    TemplateRequestContextMixin,
):
    """Adds common template variables and warning messages."""

    edc_device_app: str = "edc_device"

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        try:
            edc_device_app_config = django_apps.get_app_config(self.edc_device_app)
        except LookupError as e:
            edc_device_app_config = None
            warnings.warn(str(e))
        live_system = getattr(settings, "LIVE_SYSTEM", "TEST")
        kwargs.update(
            {
                "device_id": getattr(edc_device_app_config, "device_id", "device_id?"),
                "device_role": getattr(edc_device_app_config, "device_role", "device_role?"),
                "live_system": live_system,
            }
        )
        self.check_for_warning_messages(live_system=live_system)
        return super().get_context_data(**kwargs)

    def check_for_warning_messages(self, live_system=None) -> None:
        is_debug = getattr(settings, "DEBUG", False)

        msgs = []
        if is_debug:
            msgs.append(
                _(
                    "This EDC is running in DEBUG-mode. Use for testing only. "
                    "Do not use this system for production data collection!"
                )
            )
        elif not live_system:
            msgs.append(
                _(
                    "This EDC is for testing only. "
                    "Do not use this system for production data collection!"
                )
            )

        if self.request.user.is_superuser:
            msgs.append(
                _(
                    "You are using a `superuser` account. The EDC does not operate correctly "
                    "with user acounts that have the `superuser` status. "
                    "Update your user account before continuing."
                )
            )
        for msg in msgs:
            if msg not in [str(m) for m in messages.get_messages(self.request)]:
                messages.add_message(self.request, messages.ERROR, msg)

        warning_message = getattr(settings, "WARNING_MESSAGE", False)
        if warning_message and warning_message not in [
            str(m) for m in messages.get_messages(self.request)
        ]:
            messages.add_message(
                self.request, messages.WARNING, warning_message, extra_tags="warning"
            )



from django.apps import apps as django_apps
from django.contrib.admin import sites
from django.db import models
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch


class UrlModelMixinNoReverseMatch(Exception):  # noqa: N818
    pass


class UrlModelMixin(models.Model):
    ADMIN_SITE_NAME = None  # default is '{app_label}_admin'

    def get_absolute_url(self) -> str:
        try:
            if self.id:
                absolute_url = reverse(self.admin_url_name, args=(str(self.id),))
            else:
                absolute_url = reverse(self.admin_url_name)
        except NoReverseMatch as e:
            raise UrlModelMixinNoReverseMatch(
                f"Tried {self.admin_url_name}. Got {e}. "
                "Check urls and confirm app is in INSTALLED_APPS. If running tests, "
                "you may need to set `use_test_urls=True`."
            ) from e
        return absolute_url

    def get_changelist_url(self, search_term=None) -> str:
        url = reverse(
            f"{self.admin_site_name}:{self._meta.label_lower.replace('.', '_')}_changelist"
        )
        if search_term:
            url = f"{url}?q={search_term!s}"
        return url

    @property
    def admin_url_name(self) -> str:
        """Returns the django admin add or change url name
        (includes namespace).
        """
        mode = "change" if self.id else "add"
        app_name, model_name = self._meta.label_lower.split(".")
        return f"{self.admin_site_name}:{app_name}_{model_name}_{mode}"

    @property
    def admin_site_name(self) -> str:
        """Returns the "admin" url namespace for this model.

        Default naming convention for edc is "<app_label>_admin".
        For example, for module my_app the default would be 'my_app_admin'.

        If the edc module's admin site is not defined, defaults to "admin".
        """
        # model specific
        admin_site_name = self.ADMIN_SITE_NAME
        if not admin_site_name:
            app_label = self._meta.app_label
            try:
                # app specific
                admin_site_name = django_apps.get_app_config(app_label).admin_site_name
            except AttributeError:
                # default to edc format
                admin_site_name = f"{self._meta.app_label}_admin"
                if admin_site_name not in [s.name for s in sites.all_sites]:
                    admin_site_name = "admin"
        return admin_site_name

    @property
    def next_string(self) -> str:
        return (
            f"{self.admin_site_name}:{self._meta.label_lower.split('.')[0]}_"
            f"{self._meta.label_lower.split('.')[1]}"
        )

    class Meta:
        abstract = True



from edc_action_item.action_with_notification import ActionWithNotification
from edc_constants.constants import HIGH_PRIORITY, TBD, YES
from edc_offstudy.constants import END_OF_STUDY_ACTION

from .constants import UNBLINDING_REQUEST_ACTION, UNBLINDING_REVIEW_ACTION


class UnblindingRequestAction(ActionWithNotification):
    reference_model = "edc_unblinding.unblindingrequest"  # or inte_prn.unblindingrequest
    admin_site_name = "edc_unblinding_admin"  # or inte_prn_admin

    name = UNBLINDING_REQUEST_ACTION
    display_name = "Unblinding request"
    notification_display_name = " Unblinding request"
    parent_action_names = ()
    show_link_to_changelist = True
    show_link_to_add = True
    priority = HIGH_PRIORITY

    def get_next_actions(self):
        next_actions = []
        return self.append_to_next_if_required(
            next_actions=next_actions,
            action_name=UNBLINDING_REVIEW_ACTION,
            required=self.reference_obj.approved == TBD,
        )


class UnblindingReviewAction(ActionWithNotification):
    reference_model = "edc_unblinding.unblindingreview"  # or inte_prn.unblindingreview
    admin_site_name = "edc_unblinding_admin"  # or inte_prn_admin

    name = UNBLINDING_REVIEW_ACTION
    display_name = "Unblinding review pending"
    notification_display_name = " Unblinding review needed"
    parent_action_names = (UNBLINDING_REQUEST_ACTION,)
    show_link_to_changelist = True
    priority = HIGH_PRIORITY
    color_style = "info"
    create_by_user = False
    instructions = "This report is to be completed by the UNBLINDING REVIEWERS only."

    def get_next_actions(self):
        next_actions = []
        return self.append_to_next_if_required(
            next_actions=next_actions,
            action_name=END_OF_STUDY_ACTION,
            required=self.reference_obj.approved == YES,
        )



from __future__ import annotations

from django.contrib.auth.mixins import LoginRequiredMixin
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.views import View

from edc_appointment.utils import refresh_appointments
from edc_dashboard.url_names import url_names


class RefreshAppointmentsView(LoginRequiredMixin, View):
    onschedule_model = None

    def refresh_appointments(
        self,
        subject_identifier: str,
        visit_schedule_name: str,
        schedule_name: str,
        **kwargs,  # noqa: ARG002
    ) -> tuple[str, str]:
        return refresh_appointments(
            subject_identifier=subject_identifier,
            visit_schedule_name=visit_schedule_name,
            schedule_name=schedule_name,
            request=self.request,
        )

    def get(self, request, *args, **kwargs):  # noqa: ARG002
        subject_identifier, _ = self.refresh_appointments(**kwargs)
        url_name = url_names.get("subject_dashboard_url")
        args = (subject_identifier,)
        url = reverse(url_name, args=args)
        return HttpResponseRedirect(url)



# Generated by Django 2.1.7 on 2019-02-26 23:48

import uuid

import _socket
import django.core.validators
import django.db.models.deletion
import django_audit_fields.fields.uuid_auto_field
import django_crypto_fields.fields.encrypted_char_field
import django_crypto_fields.fields.firstname_field
import django_crypto_fields.fields.identity_field
import django_crypto_fields.fields.lastname_field
import django_revision.revision_field
import simple_history.models
from django.conf import settings
from django.db import migrations, models

import edc_model_fields.fields.date_estimated
import edc_model_fields.fields.hostname_modification_field
import edc_model_fields.fields.identity_type_field
import edc_model_fields.fields.userfield
import edc_utils
import django.utils.timezone


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("sites", "0002_alter_domain_unique"),
        ("edc_registration", "0015_auto_20181006_2257"),
    ]

    operations = [
        migrations.CreateModel(
            name="HistoricalRegisteredSubject",
            fields=[
                (
                    "created",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "modified",
                    models.DateTimeField(blank=True, default=django.utils.timezone.now),
                ),
                (
                    "user_created",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user created",
                    ),
                ),
                (
                    "user_modified",
                    edc_model_fields.fields.userfield.UserField(
                        blank=True,
                        help_text="Updated by admin.save_model",
                        max_length=50,
                        verbose_name="user modified",
                    ),
                ),
                (
                    "hostname_created",
                    models.CharField(
                        blank=True,
                        default=_socket.gethostname,
                        help_text="System field. (modified on create only)",
                        max_length=60,
                    ),
                ),
                (
                    "hostname_modified",
                    edc_model_fields.fields.hostname_modification_field.HostnameModificationField(
                        blank=True,
                        help_text="System field. (modified on every save)",
                        max_length=50,
                    ),
                ),
                (
                    "revision",
                    django_revision.revision_field.RevisionField(
                        blank=True,
                        editable=False,
                        help_text="System field. Git repository tag:branch:commit.",
                        max_length=75,
                        null=True,
                        verbose_name="Revision",
                    ),
                ),
                ("device_created", models.CharField(blank=True, max_length=10)),
                ("device_modified", models.CharField(blank=True, max_length=10)),
                (
                    "id",
                    django_audit_fields.fields.uuid_auto_field.UUIDAutoField(
                        blank=True,
                        db_index=True,
                        editable=False,
                        help_text="System auto field. UUID primary key.",
                    ),
                ),
                (
                    "subject_identifier",
                    models.CharField(
                        db_index=True, max_length=50, verbose_name="Subject Identifier"
                    ),
                ),
                (
                    "subject_identifier_as_pk",
                    models.UUIDField(default=uuid.uuid4, editable=False),
                ),
                (
                    "subject_identifier_aka",
                    models.CharField(
                        editable=False,
                        help_text="track a previously allocated identifier.",
                        max_length=50,
                        null=True,
                        verbose_name="Subject Identifier a.k.a",
                    ),
                ),
                (
                    "first_name",
                    django_crypto_fields.fields.firstname_field.FirstnameField(
                        blank=True,
                        help_text=" (Encryption: RSA local)",
                        max_length=71,
                        null=True,
                    ),
                ),
                (
                    "last_name",
                    django_crypto_fields.fields.lastname_field.LastnameField(
                        blank=True,
                        help_text=" (Encryption: RSA local)",
                        max_length=71,
                        null=True,
                        verbose_name="Last name",
                    ),
                ),
                (
                    "initials",
                    django_crypto_fields.fields.encrypted_char_field.EncryptedCharField(
                        blank=True,
                        help_text=" (Encryption: RSA local)",
                        max_length=71,
                        null=True,
                        validators=[
                            django.core.validators.RegexValidator(
                                message="Ensure initials consist of letters only in upper case, no spaces.",
                                regex="^[A-Z]{2,3}$",
                            )
                        ],
                    ),
                ),
                (
                    "dob",
                    models.DateField(
                        help_text="Format is YYYY-MM-DD",
                        null=True,
                        verbose_name="Date of birth",
                    ),
                ),
                (
                    "is_dob_estimated",
                    edc_model_fields.fields.date_estimated.IsDateEstimatedField(
                        choices=[
                            ("-", "No"),
                            ("D", "Yes, estimated the Day"),
                            ("MD", "Yes, estimated Month and Day"),
                            ("YMD", "Yes, estimated Year, Month and Day"),
                        ],
                        help_text="If the exact date is not known, please indicate which part of the date is estimated.",
                        max_length=25,
                        null=True,
                        verbose_name="Is date of birth estimated?",
                    ),
                ),
                (
                    "gender",
                    models.CharField(
                        choices=[("M", "Male"), ("F", "Female")],
                        max_length=1,
                        null=True,
                        verbose_name="Gender",
                    ),
                ),
                (
                    "subject_consent_id",
                    models.CharField(blank=True, max_length=100, null=True),
                ),
                (
                    "registration_identifier",
                    models.CharField(blank=True, max_length=36, null=True),
                ),
                (
                    "sid",
                    models.CharField(blank=True, max_length=15, null=True, verbose_name="SID"),
                ),
                (
                    "subject_type",
                    models.CharField(blank=True, max_length=25, null=True),
                ),
                (
                    "relative_identifier",
                    models.CharField(
                        blank=True,
                        help_text="For example, mother's identifier, if available / appropriate",
                        max_length=36,
                        null=True,
                        verbose_name="Identifier of immediate relation",
                    ),
                ),
                (
                    "identity",
                    django_crypto_fields.fields.identity_field.IdentityField(
                        blank=True,
                        help_text=" (Encryption: RSA local)",
                        max_length=71,
                        null=True,
                    ),
                ),
                (
                    "identity_or_pk",
                    models.CharField(
                        db_index=True,
                        default=edc_utils.get_uuid,
                        editable=False,
                        max_length=75,
                        verbose_name="identity or pk",
                    ),
                ),
                (
                    "identity_type",
                    edc_model_fields.fields.identity_type_field.IdentityTypeField(
                        blank=True,
                        choices=[
                            ("OMANG", "Omang"),
                            ("DRIVERS", "Driver's License"),
                            ("PASSPORT", "Passport"),
                            ("OMANG_RCPT", "Omang Receipt"),
                            ("OTHER", "Other"),
                        ],
                        max_length=15,
                        null=True,
                        verbose_name="What type of identity number is this?",
                    ),
                ),
                (
                    "screening_identifier",
                    models.CharField(blank=True, max_length=36, null=True),
                ),
                ("screening_datetime", models.DateTimeField(blank=True, null=True)),
                ("screening_age_in_years", models.IntegerField(blank=True, null=True)),
                ("registration_datetime", models.DateTimeField(blank=True, null=True)),
                ("randomization_datetime", models.DateTimeField(blank=True, null=True)),
                (
                    "registration_status",
                    models.CharField(
                        blank=True,
                        max_length=25,
                        null=True,
                        verbose_name="Registration status",
                    ),
                ),
                ("consent_datetime", models.DateTimeField(blank=True, null=True)),
                (
                    "comment",
                    models.TextField(
                        blank=True, max_length=250, null=True, verbose_name="Comment"
                    ),
                ),
                (
                    "additional_key",
                    models.CharField(
                        default=None,
                        editable=False,
                        help_text="A uuid (or some other text value) to be added to bypass the unique constraint of just firstname, initials, and dob.The default constraint proves limiting since the source model usually has some other attribute in additional to first_name, initials and dob which is not captured in this model",
                        max_length=36,
                        null=True,
                        verbose_name="-",
                    ),
                ),
                (
                    "dm_comment",
                    models.CharField(
                        editable=False,
                        max_length=150,
                        null=True,
                        verbose_name="Data Management comment",
                    ),
                ),
                (
                    "history_id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("history_date", models.DateTimeField()),
                ("history_change_reason", models.CharField(max_length=100, null=True)),
                (
                    "history_type",
                    models.CharField(
                        choices=[("+", "Created"), ("~", "Changed"), ("-", "Deleted")],
                        max_length=1,
                    ),
                ),
                (
                    "history_user",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "site",
                    models.ForeignKey(
                        blank=True,
                        db_constraint=False,
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        related_name="+",
                        to="sites.Site",
                    ),
                ),
            ],
            options={
                "verbose_name": "historical Registered Subject",
                "ordering": ("-history_date", "-history_id"),
                "get_latest_by": "history_date",
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        )
    ]



from __future__ import annotations

from django.contrib import admin
from django.contrib.admin import display
from django.template.loader import render_to_string
from django.urls import NoReverseMatch, reverse
from django.utils import timezone
from django_audit_fields.admin import audit_fieldset_tuple

from edc_action_item.fieldsets import action_fieldset_tuple
from edc_action_item.modeladmin_mixins import ActionItemModelAdminMixin
from edc_constants.constants import CANCELLED, OTHER
from edc_model_admin.dashboard import ModelAdminSubjectDashboardMixin
from edc_model_admin.list_filters import ReportDateListFilter
from edc_pdf_reports.admin import PdfButtonModelAdminMixin, print_selected_to_pdf_action

from ..utils import get_adverse_event_app_label
from .list_filters import CauseOfDeathListFilter, DeathDateListFilter


@admin.action(permissions=["view"], description="Print Death Reports as PDF")
def print_to_pdf_action(modeladmin, request, queryset):
    return print_selected_to_pdf_action(modeladmin, request, queryset)


class DeathReportModelAdminMixin(
    PdfButtonModelAdminMixin, ModelAdminSubjectDashboardMixin, ActionItemModelAdminMixin
):
    form = None

    add_form_template: str = "edc_adverse_event/admin/change_form.html"
    change_list_template = "edc_adverse_event/admin/change_list.html"
    change_form_template = "edc_adverse_event/admin/change_form.html"

    ordering = ("-report_datetime",)

    actions = [print_to_pdf_action]

    fieldsets = (
        (
            None,
            {
                "fields": (
                    "subject_identifier",
                    "report_datetime",
                    "death_datetime",
                    "study_day",
                    "death_as_inpatient",
                )
            },
        ),
        (
            "Opinion of Local Study Doctor",
            {"fields": ("cause_of_death", "cause_of_death_other", "narrative")},
        ),
        action_fieldset_tuple,
        audit_fieldset_tuple,
    )

    radio_fields = {
        "death_as_inpatient": admin.VERTICAL,
        "cause_of_death": admin.VERTICAL,
    }

    def get_search_fields(self, request) -> tuple[str, ...]:
        search_fields = super().get_search_fields(request)
        custom_fields = ("subject_identifier", "action_identifier")
        return tuple(set(search_fields + custom_fields))

    def get_list_display(self, request) -> tuple[str, ...]:
        list_display = super().get_list_display(request)
        custom_fields = (
            "subject_identifier",
            "dashboard",
            "pdf_button",
            "ae_button",
            "tmg_button",
            "report_datetime_with_ago",
            "death_datetime_with_ago",
            "cause_of_death_column",
            "action_item_column",
            "parent_action_item_column",
            "related_action_item_column",
        )
        return custom_fields + tuple(
            f for f in list_display if f not in custom_fields and f != "__str__"
        )

    def get_list_filter(self, request) -> tuple[str, ...]:
        list_filter = super().get_list_filter(request)
        custom_fields = (
            ReportDateListFilter,
            DeathDateListFilter,
            CauseOfDeathListFilter,
        )
        return custom_fields + tuple(f for f in list_filter if f not in custom_fields)

    @display(
        description="Cause of death",
        ordering="cause_of_death__display_name",
    )
    def cause_of_death_column(self, obj):
        cause_of_death = getattr(obj.cause_of_death, "name", obj.cause_of_death)
        if cause_of_death == OTHER:
            cause_of_death = f"Other: {obj.cause_of_death_other}"
        else:
            cause_of_death = getattr(obj.cause_of_death, "display_name", obj.cause_of_death)
        return cause_of_death

    @display(description="Report date", ordering="report_datetime")
    def report_datetime_with_ago(self, obj=None):
        return render_to_string(
            "edc_adverse_event/datetime_with_ago.html",
            dict(utc_date=timezone.now().date, report_datetime=obj.report_datetime),
        )

    @display(description="Death date", ordering="death_datetime")
    def death_datetime_with_ago(self, obj=None):
        return render_to_string(
            "edc_adverse_event/datetime_with_ago.html",
            dict(utc_date=timezone.now().date, report_datetime=obj.death_datetime),
        )

    @display(description="Action item", ordering="action_identifier")
    def action_item_column(self, obj):
        return self.get_action_item_column(obj.action_item, as_button=False)

    @display(description="Parent action", ordering="parent_action_item__action_identifier")
    def parent_action_item_column(self, obj):
        return self.get_action_item_column(obj.parent_action_item, as_button=True)

    @display(description="Related action", ordering="related_action_item__action_identifier")
    def related_action_item_column(self, obj):
        return self.get_action_item_column(obj.related_action_item, as_button=True)

    @staticmethod
    def get_action_item_column(action_item, as_button: bool | None = None):
        if action_item:
            verbose_name = action_item.reference_model_cls._meta.verbose_name
            app_label = action_item.reference_model_cls._meta.app_label
            model_name = action_item.reference_model_cls._meta.model_name
            url_name = (
                f"{get_adverse_event_app_label()}_admin:{app_label}_{model_name}_changelist"
            )
            url = reverse(url_name)
            return render_to_string(
                "edc_adverse_event/action_item_column.html",
                dict(
                    q=action_item.action_identifier,
                    url=url,
                    action_identifier=action_item.action_identifier[-9:],
                    title=verbose_name,
                    status=action_item.status,
                    status_display=action_item.get_status_display(),
                    CANCELLED=CANCELLED,
                    as_button=as_button,
                ),
            )
        return None

    @display(description="AE")
    def ae_button(self, obj):
        context = dict(
            subject_identifier=obj.subject_identifier,
            changelist_url=reverse(
                f"{self.admin_site.name}:{obj._meta.app_label}_aeinitial_changelist"
            ),
        )
        return render_to_string(
            template_name="edc_adverse_event/ae_button.html", context=context
        )

    @display(description="TMG")
    def tmg_button(self, obj):
        try:
            changelist_url = reverse(
                f"{self.admin_site.name}:{obj._meta.app_label}_deathreporttmg_changelist"
            )
        except NoReverseMatch:
            changelist_url = None
        context = dict(
            subject_identifier=obj.subject_identifier,
            changelist_url=changelist_url,
        )
        return render_to_string(
            template_name="edc_adverse_event/tmg_button.html", context=context
        )



from typing import Any

from django.db import models
from django.db.models.deletion import PROTECT

from edc_appointment.appointment_status_updater import AppointmentStatusUpdater
from edc_appointment.constants import MISSED_APPT
from edc_consent.model_mixins import RequiresConsentFieldsModelMixin
from edc_constants.constants import NO, YES
from edc_document_status.model_mixins import DocumentStatusModelMixin
from edc_identifier.model_mixins import NonUniqueSubjectIdentifierFieldMixin
from edc_metadata.model_mixins import MetadataHelperModelMixin
from edc_offstudy.model_mixins import OffstudyNonCrfModelMixin
from edc_visit_schedule.model_mixins import VisitScheduleModelMixin

from ...constants import MISSED_VISIT, NO_FOLLOW_UP_REASONS
from ...exceptions import RelatedVisitReasonError
from ...managers import VisitModelManager
from .previous_visit_model_mixin import PreviousVisitModelMixin
from .visit_model_fields_mixin import VisitModelFieldsMixin


class VisitModelMixin(
    VisitModelFieldsMixin,
    VisitScheduleModelMixin,
    NonUniqueSubjectIdentifierFieldMixin,
    PreviousVisitModelMixin,
    DocumentStatusModelMixin,
    MetadataHelperModelMixin,
    OffstudyNonCrfModelMixin,
    RequiresConsentFieldsModelMixin,
    models.Model,
):
    """
    For example:

        class SubjectVisit(VisitModelMixin, CreatesMetadataModelMixin,
            BaseUuidModel):

            class Meta(VisitModelMixin.Meta):
                app_label = 'my_app'
    """

    metadata_helper_instance_attr = None

    appointment = models.OneToOneField("edc_appointment.appointment", on_delete=PROTECT)

    objects = VisitModelManager()

    def __str__(self) -> str:
        return f"{self.subject_identifier} {self.visit_code}.{self.visit_code_sequence}"

    def save(self: Any, *args, **kwargs):
        self.subject_identifier = self.appointment.subject_identifier
        self.visit_schedule_name = self.appointment.visit_schedule_name
        self.schedule_name = self.appointment.schedule_name
        self.visit_code = self.appointment.visit_code
        self.visit_code_sequence = self.appointment.visit_code_sequence
        self.require_crfs = NO if self.reason == MISSED_VISIT else YES
        if self.appointment.appt_timing == MISSED_APPT and self.reason != MISSED_VISIT:
            raise RelatedVisitReasonError(
                "Invalid. Appointment is missed. Expected visit to be missed also. "
                f"Got visit.reason=`{self.reason}`. See {self.appointment}. "
                "Perhaps catch this in the form."
            )
        if self.appointment.appt_timing != MISSED_APPT and self.reason == MISSED_VISIT:
            raise RelatedVisitReasonError(
                "Invalid. Appointment is not missed. Did not expected a missed visit. "
                f"Got visit.reason=`{self.reason}`. See {self.appointment}. "
                "Perhaps catch this in the form."
            )
        super().save(*args, **kwargs)

    def natural_key(self) -> tuple:
        return (
            self.subject_identifier,
            self.visit_schedule_name,
            self.schedule_name,
            self.visit_code,
            self.visit_code_sequence,
        )

    natural_key.dependencies = ("edc_appointment.appointment",)  # type:ignore

    @property
    def timepoint(self) -> int:
        return self.appointment.timepoint

    @staticmethod
    def get_visit_reason_no_follow_up_choices() -> dict:
        """Returns the visit reasons that do not imply any
        data collection; that is, the subject is not available.
        """
        dct = {}
        for item in NO_FOLLOW_UP_REASONS:
            dct.update({item: item})
        return dct

    def update_appointment_status(self) -> None:
        AppointmentStatusUpdater(self.appointment)

    class Meta(NonUniqueSubjectIdentifierFieldMixin.Meta):
        abstract = True
        unique_together = (
            (
                "subject_identifier",
                "visit_schedule_name",
                "schedule_name",
                "visit_code",
                "visit_code_sequence",
            ),
            (
                "subject_identifier",
                "visit_schedule_name",
                "schedule_name",
                "report_datetime",  # implies one visit per day!
            ),
        )
        indexes = (
            models.Index(
                fields=[
                    "subject_identifier",
                    "visit_schedule_name",
                    "schedule_name",
                    "visit_code",
                    "visit_code_sequence",
                    "report_datetime",
                ]
            ),
        )



import numpy as np
import pandas as pd

from edc_constants.constants import YES

from ..mappings import yes_no
from .table import Table


class Requisition(Table):
    default_columns = [
        "subject_visit_id",
        "requisition_identifier",
        "panel_name",
        "requisition_datetime",
        "is_drawn",
        "reason_not_drawn",
        "drawn_datetime",
        "specimen_type",
        "study_site_name",
    ]

    def prepare(self, df_visit=None, **kwargs):
        self.dataframe["requisition_datetime"] = self.helper.to_local_datetime(
            self.dataframe["requisition_datetime"]
        )
        self.dataframe["drawn_datetime"] = self.helper.to_local_datetime(
            self.dataframe["drawn_datetime"]
        )
        self.dataframe["received_datetime"] = self.helper.to_local_datetime(
            self.dataframe["received_datetime"]
        )
        self.dataframe["received"] = self.dataframe["received"].map(yes_no, na_action="ignore")
        self.dataframe["processed"] = self.dataframe["processed"].map(
            yes_no, na_action="ignore"
        )
        self.dataframe["packed"] = self.dataframe["packed"].map(yes_no, na_action="ignore")
        self.dataframe["shipped"] = self.dataframe["shipped"].map(yes_no, na_action="ignore")
        self.dataframe = pd.merge(self.dataframe, df_visit, on="subject_visit_id", how="left")
        self.dataframe = self.dataframe.drop(["subject_visit_id"], axis=1)
        # fix reason_not_drawn
        is_blank = (self.dataframe.reason_not_drawn.notnull()) & (
            self.dataframe.is_drawn == YES
        )
        self.dataframe.loc[is_blank, "reason_not_drawn"] = np.nan
        self.dataframe = self.dataframe[self.reordered_columns]

    @property
    def reordered_columns(self):
        columns = [
            col
            for col in list(self.dataframe.columns)
            if col
            not in [
                "subject_identifier",
                "gender",
                "dob",
                "visit_code",
                "visit_datetime",
                "visit_date",
            ]
        ]
        columns.insert(0, "visit_code")
        columns.insert(0, "visit_date")
        columns.insert(0, "visit_datetime")
        columns.insert(0, "dob")
        columns.insert(0, "gender")
        columns.insert(0, "subject_identifier")
        return columns



from edc_list_data.model_mixins import ListModelMixin


class ActionsRequired(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "Actions Required"
        verbose_name_plural = "Actions Required"


class ProtocolViolations(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "Protocol Violations"
        verbose_name_plural = "Protocol Violations"


class ProtocolIncidents(ListModelMixin):
    class Meta(ListModelMixin.Meta):
        verbose_name = "Protocol Incidents"
        verbose_name_plural = "Protocol Incidents"



import sys

from django.apps import AppConfig as DjangoAppConfig
from django.core.checks import register
from django.core.management.color import color_style

from .research_protocol_config import ResearchProtocolConfig
from .system_checks import middleware_check

style = color_style()


class AppConfig(DjangoAppConfig):
    name = "edc_protocol"
    verbose_name = "Edc Protocol"
    include_in_administration_section = True
    messages_written = False

    def ready(self):
        register(middleware_check)
        sys.stdout.write(f"Loading {self.verbose_name} ...\n")
        protocol = ResearchProtocolConfig()
        sys.stdout.write(f" * {protocol.protocol}: {protocol.protocol_name}.\n")
        open_date = protocol.study_open_datetime.strftime("%Y-%m-%d %Z")
        sys.stdout.write(f" * Study opening date: {open_date}\n")
        close_date = protocol.study_close_datetime.strftime("%Y-%m-%d %Z")
        sys.stdout.write(f" * Expected study closing date: {close_date}\n")
        sys.stdout.write(f" Done loading {self.verbose_name}.\n")
        sys.stdout.flush()
        self.messages_written = True



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING
from uuid import uuid4

from django.core.handlers.wsgi import WSGIRequest
from django.utils.translation import gettext as _

from .next_querystring import NextQuerystring
from .perms import Perms

if TYPE_CHECKING:
    from django.contrib.auth.models import User
    from django.contrib.sites.models import Site

    from edc_model.models import BaseUuidModel

    class Model(BaseUuidModel):
        subject_identifier: str
        ...

    class WSGIRequestObject(WSGIRequest):
        site: Site


ADD: int = 0
CHANGE = 1
VIEW = 2

__all__ = ["ADD", "CHANGE", "VIEW", "ModelButton"]


class ModelButtonError(Exception):
    pass


@dataclass(kw_only=True)
class ModelButton:
    user: User = None
    model_obj: Model = None
    model_cls: type[Model] = field(default=None)
    current_site: Site = None
    subject_identifier: str | None = None
    request: WSGIRequestObject | None = None
    fixed_label: str | None = None
    labels: tuple[str, str, str] = field(default=("Add", "Change", "View"))
    fa_icons: tuple[str, str, str] = field(default=("fas fa-plus", "fas fa-pen", "fas fa-eye"))
    fixed_color: str | None = None
    colors: tuple[str, str, str] = field(default=("warning", "success", "default"))
    titles: tuple[str, str, str] = field(default=("Add", "Change", "View only"))
    next_url_name: str = field(default=None)
    _action: int = field(default=None, init=False)
    _perms: Perms = field(default=None, init=False)

    def __post_init__(self):
        if self.model_obj:
            self.model_cls = self.model_obj._meta.model
        if not self.model_cls:
            raise ModelButtonError(f"Model class is required if instance=None. See {self}.")

    @property
    def fa_icon(self) -> str:
        return self.fa_icons[self.action]

    @property
    def label(self) -> str:
        if self.fixed_label:
            return _(self.fixed_label)
        return _(self.labels[self.action])

    @property
    def color(self) -> str:
        if self.fixed_color:
            return self.fixed_color
        return self.colors[self.action]

    @property
    def title(self) -> str:
        return self.titles[self.action]

    @property
    def action(self):
        if not self._action:
            self._action = VIEW
            if not self.model_obj:
                self._action = ADD
            elif self.model_obj and self.perms.change:
                self._action = CHANGE
        return self._action

    @property
    def perms(self) -> Perms:
        if not self._perms:
            self._perms = Perms(
                model_cls=self.model_cls,
                user=self.user,
                current_site=self.current_site,
                site=self.site,
            )
        return self._perms

    @property
    def disabled(self) -> str:
        disabled = "disabled"
        if (not self.model_obj and self.perms.add) or (
            self.model_obj and (self.perms.change or self.perms.view)
        ):
            disabled = ""
        return disabled

    @property
    def btn_id(self) -> str:
        btn_id = f"{self.model_cls._meta.label_lower.split('.')[1]}-{uuid4().hex}"
        if self.model_obj:
            btn_id = (
                f"{self.model_cls._meta.label_lower.split('.')[1]}-{self.model_obj.id.hex}"
            )
        return btn_id

    @property
    def site(self) -> Site | None:
        """If model_obj is None, then Site should come from the
        request object (if add).
        """
        return getattr(self.model_obj, "site", None) or getattr(self.request, "site", None)

    @property
    def url(self) -> str:
        if self.action == ADD:
            url = "?".join([f"{self.model_cls().get_absolute_url()}", self.querystring])
        else:
            url = "?".join([f"{self.model_obj.get_absolute_url()}", self.querystring])
        return url

    @property
    def querystring(self) -> str:
        nq = NextQuerystring(
            next_url_name=self.next_url_name,
            reverse_kwargs=self.reverse_kwargs,
            extra_kwargs=self.extra_kwargs,
        )
        return nq.querystring

    @property
    def reverse_kwargs(self) -> dict[str, str]:
        return dict(
            subject_identifier=self.subject_identifier or self.model_obj.subject_identifier,
        )

    @property
    def extra_kwargs(self) -> dict[str, str | int]:
        return {}



from django.db import models

from edc_reportable.units import PLUS

from ..model_mixin_factories import reportable_result_model_mixin_factory


class ProteinuriaModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="proteinuria",
        verbose_name="Proteinuria",
        units_choices=((PLUS, PLUS),),
        decimal_places=1,
    ),
    models.Model,
):
    class Meta:
        abstract = True



# Generated by Django 5.1.2 on 2024-11-15 14:54

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        (
            "edc_pharmacy",
            "0027_rename_at_location_historicalstock_transferred_and_more",
        ),
    ]

    operations = [
        migrations.RemoveField(
            model_name="historicalstock",
            name="transferred",
        ),
        migrations.RemoveField(
            model_name="stock",
            name="transferred",
        ),
    ]



from __future__ import annotations

from datetime import date

from django.db import models

from edc_constants.constants import NO, NOT_APPLICABLE, YES
from edc_model import duration_to_date


def calculate_date(
    instance: models.Model, fld_prefix: str, reference_field: str
) -> tuple[date | None, str]:
    """Returns tuple date and YES/NO/NA.

    The date is either the actual date or an estimated date
    based on the ago field value.
    ."""
    calculated_date: date | None = None
    is_estimated: str = NOT_APPLICABLE
    report_datetime = getattr(instance, reference_field)
    rx_date = getattr(instance, f"{fld_prefix}_date")
    rx_ago = getattr(instance, f"{fld_prefix}_ago")
    if rx_ago and not rx_date and report_datetime:
        if rx_ago and not rx_date:
            calculated_date = duration_to_date(rx_ago, report_datetime)
            is_estimated = YES
        elif rx_date:
            calculated_date = rx_date
            is_estimated = NO

        else:
            calculated_date = None
            is_estimated = NOT_APPLICABLE
    return calculated_date, is_estimated


def update_calculated_date(instance: models.Model, fld_prefix: str, reference_field: str):
    """Wrapper for model save method"""
    calculated_date, is_calculated = calculate_date(
        instance=instance, fld_prefix=fld_prefix, reference_field=reference_field
    )
    setattr(instance, f"{fld_prefix}_calculated_date", calculated_date)
    setattr(instance, f"{fld_prefix}_date_is_estimated", is_calculated)



from .crf_df_handler import CrfDfHandler


class NonCrfDfHandler(CrfDfHandler):
    sort_by = None

    def prepare_dataframe(self, **kwargs):
        columns = [col for col in self.columns if col in self.dataframe.columns]
        self.dataframe = self.dataframe.loc[:, columns]

    @property
    def columns(self):
        """Returns a list of column names."""
        columns = list(self.dataframe.columns)
        # "export_" columns
        if self.exclude_export_columns:
            columns = [col for col in columns if not col.startswith("export_")]
        # "system" columns, move to the end
        if not self.exclude_system_columns:
            columns = [col for col in columns if col not in self.system_columns]
            columns.extend(self.system_columns)
        return columns



from django.core.management.base import BaseCommand

from edc_auth.fix_export_permissions import ExportPermissionsFixer


class Command(BaseCommand):
    def handle(self, *args, **options):
        fixer = ExportPermissionsFixer()
        fixer.fix()



# Generated by Django 4.2.7 on 2023-12-06 16:31

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("edc_lab", "0032_alter_consignee_options_and_more"),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name="panel",
            name="edc_lab_pan_lab_pro_2a0c6d_idx",
        ),
        migrations.AddIndex(
            model_name="panel",
            index=models.Index(
                fields=["modified", "created"], name="edc_lab_pan_modifie_a465ed_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="panel",
            index=models.Index(
                fields=["user_modified", "user_created"],
                name="edc_lab_pan_user_mo_ff00b2_idx",
            ),
        ),
    ]



from django.db import models
from django.utils import timezone
from sequences import get_next_value

from edc_model.models import BaseUuidModel, HistoricalRecords

from ...exceptions import StockTransferError
from .location import Location


class Manager(models.Manager):
    use_in_migrations = True


class StockTransfer(BaseUuidModel):
    """A model to track allocated stock transfers from location A
    to location B.
    """

    transfer_identifier = models.CharField(
        max_length=36,
        unique=True,
        null=True,
        blank=True,
        help_text="A sequential unique identifier set by the EDC",
    )

    transfer_datetime = models.DateTimeField(default=timezone.now)

    from_location = models.ForeignKey(
        Location,
        on_delete=models.PROTECT,
        null=True,
        blank=False,
        related_name="from_location",
        limit_choices_to={"site__isnull": True},
    )
    to_location = models.ForeignKey(
        Location,
        on_delete=models.PROTECT,
        null=True,
        blank=False,
        related_name="to_location",
        limit_choices_to={"site__isnull": False},
    )

    item_count = models.PositiveIntegerField(null=True, blank=False)

    objects = Manager()

    history = HistoricalRecords()

    def __str__(self):
        return self.transfer_identifier

    def save(self, *args, **kwargs):
        if not self.transfer_identifier:
            self.transfer_identifier = f"{get_next_value(self._meta.label_lower):06d}"
            if self.from_location == self.to_location:
                raise StockTransferError("Locations cannot be the same")
        super().save(*args, **kwargs)

    @property
    def comfirmed_items(self) -> int:
        return self.stocktransferitem_set.filter(
            stock__confirmationatsiteitem__isnull=False
        ).count()

    @property
    def uncomfirmed_items(self) -> int:
        return self.stocktransferitem_set.filter(
            stock__confirmationatsiteitem__isnull=True
        ).count()

    @property
    def export_references(self):
        return "Export ref"

    @property
    def shipped(self):
        return False

    @property
    def export_datetime(self):
        return timezone.now()

    @property
    def site(self):
        class Dummy:
            name = "sitename"

        return Dummy()

    @property
    def consignee(self):
        class Dummy:
            country = "Tanzania"

        return Dummy()

    @property
    def shipper(self):
        return {
            "contact_name": "Deus Buma",
            "name": "META III Central Pharmacy",
            "address": "",
            "city": "",
            "state": "",
            "postal_code": "0000",
            "country": "Tanzania",
        }

    class Meta(BaseUuidModel.Meta):
        verbose_name = "Stock transfer"
        verbose_name_plural = "Stock transfers"



# Generated by Django 5.2.3 on 2025-06-26 22:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0069_alter_historicalstoragebin_capacity_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="stockout",
            name="codes",
            field=models.CharField(max_length=150, null=True),
        ),
    ]



import sys

from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand

from edc_action_item.models import ActionItem
from edc_action_item.site_action_items import site_action_items


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "--dry-run",
            action="store_true",
            dest="dry-run",
            default=False,
            help="dry run",
        )

    def handle(self, *args, **options):
        self.dry_run = options["dry-run"]
        if self.dry_run:
            sys.stdout.write(self.style.NOTICE("\nDry run. No changes will be made.\n"))

        actions = []
        for action in site_action_items.registry.values():
            if action.related_reference_fk_attr:
                actions.append(action)

        print("Ready to inspect actions with a `related_reference_fk_attr`")
        names = [a.name for a in actions]
        print("Inspecting action types:")
        for name in names:
            print(f"  * {name}")

        opts = dict(
            related_action_item__isnull=True,
            action_type__name__in=[a.name for a in actions],
        )

        qs = ActionItem.objects.filter(**opts).order_by("subject_identifier", "created")
        print(f"Found {qs.count()} action_items with missing `related_action_item`.")
        self.inspect_action_items(**opts)
        print("Done.")

    def inspect_action_items(self, **opts):
        for action_item in ActionItem.objects.filter(**opts).order_by(
            "subject_identifier", "created"
        ):
            print(
                action_item,
                action_item.action_cls.related_reference_fk_attr,
                action_item.parent_action_item,
            )
            model_cls = django_apps.get_model(action_item.action_cls.get_reference_model())
            try:
                obj = model_cls.objects.get(action_identifier=action_item.action_identifier)
            except ObjectDoesNotExist:
                print("  skipping")
            else:
                related_obj = getattr(obj, action_item.action_cls.related_reference_fk_attr)
                if not self.dry_run:
                    related_obj.save()
                else:
                    print("  not saving")
                related_action_item = ActionItem.objects.get(
                    action_identifier=related_obj.action_identifier
                )
                action_item.related_action_item = related_action_item
                if not self.dry_run:
                    action_item.save()
                else:
                    print("  not saving")
                print(f"  - updated, set related_action_item={related_action_item}")
            if not self.dry_run:
                model_cls.save()
            else:
                print("  not saving")
            print("******************************************************")



from django.contrib import admin

from ..admin_site import edc_qareports_admin
from ..forms import NoteForm
from ..modeladmin_mixins import NoteModelAdminMixin
from ..models import Note


@admin.register(Note, site=edc_qareports_admin)
class NoteModelAdmin(
    NoteModelAdminMixin,
    admin.ModelAdmin,
):
    """A modeladmin class for the Note model."""

    form = NoteForm
    note_template_name = "edc_qareports/qa_report_note.html"



from django.apps import apps as django_apps

from edc_constants.constants import NEW
from edc_metadata import REQUIRED

from ..constants import COMPLETE_APPT, IN_PROGRESS_APPT, INCOMPLETE_APPT, NEW_APPT


def appointment_mark_as_done(modeladmin, request, queryset):
    """Update appointment to DONE.

    If a record exists in CrfStatus, set to INCOMPLETE.
    """
    crfstatus_model_cls = django_apps.get_model("edc_crf.crfstatus")
    crfmetadata_model_cls = django_apps.get_model("edc_metadata.crfmetadata")

    for obj in queryset.filter(
        appt_status__in=[COMPLETE_APPT, INCOMPLETE_APPT, IN_PROGRESS_APPT]
    ):
        if crfstatus_model_cls.objects.filter(
            subject_identifier=obj.subject_identifier,
            visit_schedule_name=obj.visit_schedule_name,
            schedule_name=obj.schedule_name,
            visit_code=obj.visit_code,
            visit_code_sequence=obj.visit_code_sequence,
        ).exists():
            obj.appt_status = INCOMPLETE_APPT
            obj.save(update_fields=["appt_status"])
        else:
            if crfmetadata_model_cls.objects.filter(
                subject_identifier=obj.subject_identifier,
                visit_schedule_name=obj.visit_schedule_name,
                schedule_name=obj.schedule_name,
                visit_code=obj.visit_code,
                visit_code_sequence=obj.visit_code_sequence,
                entry_status=REQUIRED,
            ).exists():
                obj.appt_status = INCOMPLETE_APPT
            else:
                obj.appt_status = COMPLETE_APPT
            obj.save(update_fields=["appt_status"])


# noinspection PyTypeHints
appointment_mark_as_done.short_description = "Mark as DONE (if allowed)"  # type: ignore


def appointment_mark_as_new(modeladmin, request, queryset):
    """Update appointment to NEW."""
    crfmetadata_model_cls = django_apps.get_model("edc_metadata.crfmetadata")
    for obj in queryset.filter(appt_status__in=[NEW, COMPLETE_APPT]):
        if not crfmetadata_model_cls.objects.filter(
            subject_identifier=obj.subject_identifier,
            visit_schedule_name=obj.visit_schedule_name,
            schedule_name=obj.schedule_name,
            visit_code=obj.visit_code,
            visit_code_sequence=obj.visit_code_sequence,
        ).exists():
            obj.appt_status = NEW_APPT
            obj.save(update_fields=["appt_status"])


# noinspection PyTypeHints
appointment_mark_as_new.short_description = "Mark as NEW (if allowed)"  # type: ignore



from edc_constants.constants import NOT_APPLICABLE, OTHER

from .constants import DIET_LIFESTYLE, DRUGS, INSULIN, THIS_CLINIC

CARE_ACCESS = (
    (THIS_CLINIC, "Patient comes to this facility for their care"),
    (OTHER, "Patient goes to a different clinic"),
    (NOT_APPLICABLE, "Not applicable"),
)

CHOL_MANAGEMENT = (
    (DRUGS, "Oral drugs"),
    (DIET_LIFESTYLE, "Diet and lifestyle alone"),
)
DM_MANAGEMENT = (
    (INSULIN, "Insulin injections"),
    (DRUGS, "Oral drugs"),
    (DIET_LIFESTYLE, "Diet and lifestyle alone"),
)

HTN_MANAGEMENT = (
    (DRUGS, "Drugs / Medicine"),
    (DIET_LIFESTYLE, "Diet and lifestyle alone"),
)



from django.db import models

from edc_qareports.model_mixins import QaReportModelMixin, qa_reports_permissions


class StockAvailability(QaReportModelMixin, models.Model):
    subject_identifier = models.CharField(max_length=50, default="")

    visit_code = models.DecimalField(max_digits=8, decimal_places=1, null=True)

    appt_date = models.DateField(null=True)

    baseline_date = models.DateField(null=True)

    relative_days = models.IntegerField(null=True)

    codes = models.TextField(default="")

    bins = models.TextField(default="")

    class Meta(QaReportModelMixin.Meta):
        verbose_name = "Stock availability"
        verbose_name_plural = "Stock availability"
        default_permissions = qa_reports_permissions



from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist

from edc_constants.constants import NEW

from .get_action_type import get_action_type


class ActionItemDeleteError(Exception):
    pass


def delete_action_item(action_cls=None, subject_identifier=None, using=None):
    """Deletes any NEW action items for a given class
    and subject_identifier.
    """
    try:
        obj = (
            action_cls.action_item_model_cls()
            .objects.using(using)
            .get(
                subject_identifier=subject_identifier,
                action_type=get_action_type(action_cls),
                status=NEW,
            )
        )
    except ObjectDoesNotExist:
        raise ActionItemDeleteError(
            "Unable to delete action item. "
            f"Action item {action_cls.name} does not exist for "
            f"{subject_identifier}."
        )
    except MultipleObjectsReturned:
        for obj in (
            action_cls.action_item_model_cls()
            .objects.using(using)
            .filter(
                subject_identifier=subject_identifier,
                action_type=get_action_type(action_cls),
                status=NEW,
            )
        ):
            obj.delete(using=using)
    else:
        obj.delete(using=using)



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from dateutil.relativedelta import relativedelta
from django.db import transaction

from ..constants import INCOMPLETE_APPT, NEW_APPT
from ..utils import get_appointment_by_datetime
from .unscheduled_appointment_creator import (
    CreateAppointmentError,
    UnscheduledAppointmentCreator,
)

if TYPE_CHECKING:
    from ..models import Appointment


def create_next_appointment_as_interim(
    next_appt_datetime: datetime = None, appointment: Appointment = None
) -> Appointment | None:
    return create_unscheduled_appointment(next_appt_datetime, appointment)


def create_unscheduled_appointment(
    next_appt_datetime: datetime = None, appointment: Appointment = None
) -> Appointment | None:
    """Create and return an unscheduled appointment if
    `next_appt_datetime` is within the window period of
    `appointment`.

    """
    unscheduled_appointment = None
    next_appt_datetime = next_appt_datetime or appointment.appt_datetime + relativedelta(
        days=1
    )
    next_appointment = get_appointment_by_datetime(
        next_appt_datetime,
        appointment.subject_identifier,
        appointment.visit_schedule_name,
        appointment.schedule_name,
        raise_if_in_gap=False,
    )
    if appointment == next_appointment:
        if (
            appointment.relative_next
            and appointment.relative_next.appt_status == NEW_APPT
            and appointment.relative_next.visit_code_sequence > 0
        ):
            appointment.relative_next.delete()
        appt_status = appointment.appt_status
        appointment.appt_status = INCOMPLETE_APPT
        appointment.save_base(update_fields=["appt_status"])
        with transaction.atomic():
            try:
                unscheduled_appointment = UnscheduledAppointmentCreator(
                    subject_identifier=appointment.subject_identifier,
                    visit_schedule_name=appointment.visit_schedule_name,
                    schedule_name=appointment.schedule_name,
                    visit_code=appointment.visit_code,
                    facility=appointment.facility,
                    suggested_appt_datetime=next_appt_datetime,
                    suggested_visit_code_sequence=appointment.visit_code_sequence + 1,
                )
            except CreateAppointmentError:
                pass
        appointment.appt_status = appt_status
        appointment.save_base(update_fields=["appt_status"])
    return unscheduled_appointment



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING
from zoneinfo import ZoneInfo

from django.core.exceptions import ObjectDoesNotExist

from edc_constants.constants import NO, YES
from edc_crf.crf_form_validator import CrfFormValidator
from edc_form_validators import INVALID_ERROR
from edc_utils import formatted_datetime

from ...utils import get_rx_model_cls, get_rxrefill_model_cls

if TYPE_CHECKING:
    from edc_appointment.models import Appointment

    from ...models import Formulation, Medication, Rx, RxRefill


class StudyMedicationFormValidator(CrfFormValidator):
    def clean(self):
        next_appt_datetime = None

        self.confirm_has_rx()

        if self.related_visit.appointment.relative_next:
            next_appt_datetime = self.related_visit.appointment.relative_next.appt_datetime

        if next_appt_datetime:
            if (
                self.refill_start_datetime
                and self.refill_start_datetime.astimezone(ZoneInfo("UTC")) > next_appt_datetime
            ):
                local_dte = formatted_datetime(next_appt_datetime)
                error_msg = (
                    "Refill start date cannot be after next appointmnent date. "
                    f"Next appointment date is {local_dte}."
                )

                self.raise_validation_error(
                    {"refill_start_datetime": error_msg}, INVALID_ERROR
                )

        self.validate_refill_start_date_against_offschedule_date()

        self.required_if(
            NO,
            field="refill_to_next_visit",
            field_required="refill_end_datetime",
            inverse=False,
        )

        if not self.next_appointment and self.cleaned_data.get("refill_to_next_visit") == YES:
            error_msg = "Invalid. Subject does not have a future appointment."
            self.raise_validation_error({"refill_to_next_visit": error_msg}, INVALID_ERROR)

        self.validate_refill_dates()

        self.validate_refill_end_date_against_offschedule_date()

    def confirm_has_rx(self) -> Rx | None:
        return self.rx

    @property
    def refill_start_datetime(self) -> datetime | None:
        if refill_start_datetime := self.cleaned_data.get("refill_start_datetime"):
            return refill_start_datetime.astimezone(ZoneInfo("UTC"))
        return refill_start_datetime

    @property
    def refill_end_datetime(self) -> datetime | None:
        if refill_end_datetime := self.cleaned_data.get("refill_end_datetime"):
            return refill_end_datetime.astimezone(ZoneInfo("UTC"))
        return refill_end_datetime

    @property
    def next_refill(self) -> RxRefill | None:
        next_refill = None
        if self.refill_start_datetime:
            for obj in (
                get_rxrefill_model_cls()
                .objects.filter(
                    rx=self.rx,
                    refill_start_datetime__gt=self.refill_start_datetime,
                )
                .order_by("refill_start_datetime")
            ):
                next_refill = obj
        return next_refill

    @property
    def rx(self) -> Rx:
        try:
            return get_rx_model_cls().objects.get(
                subject_identifier=self.subject_identifier,
                medications__in=[self.medication],
            )
        except ObjectDoesNotExist:
            self.raise_validation_error(
                {"__all__": "Prescription does not exist"}, INVALID_ERROR
            )

    @property
    def formulation(self) -> Formulation | None:
        return self.cleaned_data.get("formulation")

    @property
    def medication(self) -> Medication:
        if self.formulation:
            return self.formulation.medication
        self.raise_validation_error(
            {"__all__": "Need the formulation to look up the prescription."},
            INVALID_ERROR,
        )

    @property
    def next_appointment(self) -> Appointment | None:
        return self.related_visit.appointment.next

    def validate_refill_start_date_against_offschedule_date(self):
        if (
            self.offschedule_datetime
            and self.refill_start_datetime
            and self.refill_start_datetime > self.offschedule_datetime
        ):
            error_msg = (
                "Invalid. Cannot be after offschedule datetime. "
                f"Got {self.offschedule_datetime}."
            )
            self.raise_validation_error({"refill_start_datetime": error_msg}, INVALID_ERROR)

    def validate_refill_end_date_against_offschedule_date(self):
        if (
            self.offschedule_datetime
            and self.refill_end_datetime
            and self.refill_end_datetime > self.offschedule_datetime
        ):
            error_msg = (
                "Invalid. Cannot be after offschedule datetime. "
                f"Got {self.offschedule_datetime}."
            )
            self.raise_validation_error({"refill_end_datetime": error_msg}, INVALID_ERROR)

    def validate_refill_dates(self):
        if (
            self.refill_start_datetime
            and self.cleaned_data.get("refill")
            and self.cleaned_data.get("refill") == NO
        ):
            error_msg = (
                "Subject is not receiving study medication. Refill end date "
                "and time must exactly match refill start date and time."
            )
            if (
                not self.refill_end_datetime
                or self.refill_start_datetime != self.refill_end_datetime
            ):
                self.raise_validation_error({"refill_end_datetime": error_msg}, INVALID_ERROR)

        if (
            self.cleaned_data.get("refill")
            and self.refill_start_datetime
            and self.refill_end_datetime
            and self.cleaned_data.get("refill") == YES
            and self.refill_start_datetime >= self.refill_end_datetime
        ):
            if self.cleaned_data.get("refill_to_next_visit") == YES:
                error_msg = (
                    "Invalid. The calculated refill end date will be before "
                    f"the start date! Got {self.cleaned_data.get('refill_end_datetime')}. "
                    "Check the refill start date."
                )
            else:
                error_msg = (
                    "Invalid. Refill end date must be after the refill start date and "
                    "before the next visit"
                )
            self.raise_validation_error({"refill_end_datetime": error_msg}, INVALID_ERROR)



from django.apps import apps as django_apps
from django.contrib import admin
from django.core.exceptions import ObjectDoesNotExist
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.translation import gettext as _

from edc_model_admin.dashboard import ModelAdminDashboardMixin
from edc_model_admin.mixins import TemplatesModelAdminMixin
from edc_sites.admin import SiteModelAdminMixin
from edc_visit_schedule.admin import ScheduleStatusListFilter
from edc_visit_tracking.utils import get_related_visit_model_cls

from .qa_report_modeladmin_mixin import QaReportModelAdminMixin


class OnStudyMissingValuesModelAdminMixin(
    QaReportModelAdminMixin,
    SiteModelAdminMixin,
    ModelAdminDashboardMixin,
    TemplatesModelAdminMixin,
):
    include_note_column: bool = True
    site_list_display_insert_pos: int = 2
    qa_report_list_display_insert_pos = 4
    ordering = ["site", "subject_identifier"]

    list_display = [
        "dashboard",
        "render_button",
        "subject",
        "site",
        "label",
        "crf",
        "visit",
        "report_date",
        "created",
    ]

    list_filter = [
        ScheduleStatusListFilter,
        "label",
        "visit_code",
        "report_datetime",
    ]

    search_fields = ["subject_identifier", "label"]

    @admin.display(description="Update")
    def render_button(self, obj=None):
        crf_model_cls = django_apps.get_model(obj.label_lower)
        try:
            django_apps.get_model(obj.label_lower).objects.get(id=obj.original_id)
        except ObjectDoesNotExist:
            url = reverse(
                f"{self.crf_admin_site_name(crf_model_cls)}:"
                f"{obj.label_lower.replace('.', '_')}_add"
            )
            url = (
                f"{url}?next={self.admin_site.name}:"
                f"{self.model._meta.label_lower.replace('.', '_')}_changelist"
                f"&subject_identifier={obj.subject_identifier}"
                f"&subject_visit={obj.subject_visit_id}"
                f"&appointment={self.related_visit(obj).appointment.id}"
                f"&requisition={obj.original_id}"
            )
            title = _(f"Add {crf_model_cls._meta.verbose_name}")
            label = _("Add CRF")
            crf_button = render_to_string(
                "edc_qareports/columns/add_button.html",
                context=dict(title=title, url=url, label=label),
            )
        else:
            url = reverse(
                f"{self.crf_admin_site_name(crf_model_cls)}:"
                f"{obj.label_lower.replace('.', '_')}_change",
                args=(obj.original_id,),
            )
            url = (
                f"{url}?next={self.admin_site.name}:"
                f"{self.model._meta.label_lower.replace('.', '_')}_changelist"
            )
            title = _(f"Change {crf_model_cls._meta.verbose_name}")
            label = _("Change CRF")
            crf_button = render_to_string(
                "edc_qareports/columns/change_button.html",
                context=dict(title=title, url=url, label=label),
            )
        return crf_button

    def dashboard(self, obj=None, label=None) -> str:
        dashboard_url = reverse(
            self.get_subject_dashboard_url_name(obj=obj),
            kwargs=dict(
                subject_identifier=obj.subject_identifier,
                appointment=self.related_visit(obj).appointment.id,
            ),
        )
        context = dict(title=_("Go to subject's dashboard"), url=dashboard_url, label=label)
        return render_to_string("edc_subject_dashboard/dashboard_button.html", context=context)

    @staticmethod
    def crf_admin_site_name(crf_model_cls) -> str:
        """Returns the name of the admin site CRFs are registered
        by assuming admin site name follows the edc naming convention.

        For example: 'meta_subject_admin' or 'effect_subject_admin'
        """
        return f"{crf_model_cls._meta.label_lower.split('.')[0]}_admin"

    @staticmethod
    def related_visit(obj=None):
        return get_related_visit_model_cls().objects.get(id=obj.subject_visit_id)

    @admin.display(description="CRF", ordering="label_lower")
    def crf(self, obj=None) -> str:
        model_cls = django_apps.get_model(obj.label_lower)
        return model_cls._meta.verbose_name

    @admin.display(description="Visit", ordering="visit_code")
    def visit(self, obj=None) -> str:
        return f"{obj.visit_code}.{obj.visit_code_sequence}"

    @admin.display(description="Report date", ordering="report_datetime")
    def report_date(self, obj) -> str | None:
        if obj.report_datetime:
            return obj.report_datetime.date()
        return None

    @admin.display(description="Subject", ordering="subject_identifier")
    def subject(self, obj) -> str | None:
        url = reverse(
            f"{self.admin_site.name}:{self.model._meta.label_lower.replace('.', '_')}"
            "_changelist"
        )
        return render_to_string(
            "edc_qareports/columns/subject_identifier_column.html",
            {
                "subject_identifier": obj.subject_identifier,
                "url": url,
                "title": _("Filter by subject"),
            },
        )



# Generated by Django 6.0 on 2025-03-12 10:41

import django.db.models.deletion
import django.db.models.manager
from django.db import migrations, models

import edc_pharmacy.models.stock.allocation
import edc_pharmacy.models.stock.confirmation_at_site
import edc_pharmacy.models.stock.confirmation_at_site_item
import edc_pharmacy.models.stock.dispense
import edc_pharmacy.models.stock.dispense_item
import edc_sites.managers


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0060_alter_container_max_per_subject_and_more"),
        ("sites", "0002_alter_domain_unique"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="historicalstocktransferconfirmation",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical Stock Transfer Site Confirmation",
                "verbose_name_plural": "historical Stock Transfer Site Confirmations",
            },
        ),
        migrations.AlterModelOptions(
            name="stocktransferconfirmation",
            options={
                "default_manager_name": "objects",
                "default_permissions": (
                    "add",
                    "change",
                    "delete",
                    "view",
                    "export",
                    "import",
                ),
                "verbose_name": "Stock Transfer Site Confirmation",
                "verbose_name_plural": "Stock Transfer Site Confirmations",
            },
        ),
        migrations.AlterModelManagers(
            name="allocation",
            managers=[
                ("objects", edc_pharmacy.models.stock.allocation.Manager()),
                ("on_site", edc_sites.managers.CurrentSiteManager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="dispense",
            managers=[
                ("objects", edc_pharmacy.models.stock.dispense.Manager()),
                ("on_site", edc_sites.managers.CurrentSiteManager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="dispenseitem",
            managers=[
                ("objects", edc_pharmacy.models.stock.dispense_item.Manager()),
                ("on_site", edc_sites.managers.CurrentSiteManager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="stocktransferconfirmation",
            managers=[
                ("objects", edc_pharmacy.models.stock.confirmation_at_site.Manager()),
                ("on_site", edc_sites.managers.CurrentSiteManager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="stocktransferconfirmationitem",
            managers=[
                (
                    "objects",
                    edc_pharmacy.models.stock.confirmation_at_site_item.Manager(),
                ),
                ("on_site", edc_sites.managers.CurrentSiteManager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="storagebin",
            managers=[
                ("on_site", edc_sites.managers.CurrentSiteManager()),
                ("objects", django.db.models.manager.Manager()),
            ],
        ),
        migrations.AlterModelManagers(
            name="storagebinitem",
            managers=[
                ("on_site", edc_sites.managers.CurrentSiteManager()),
                ("objects", django.db.models.manager.Manager()),
            ],
        ),
        migrations.AddField(
            model_name="allocation",
            name="site",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="sites.site",
            ),
        ),
        migrations.AddField(
            model_name="dispense",
            name="site",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="sites.site",
            ),
        ),
        migrations.AddField(
            model_name="dispenseitem",
            name="site",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="sites.site",
            ),
        ),
        migrations.AddField(
            model_name="historicalallocation",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.site",
            ),
        ),
        migrations.AddField(
            model_name="historicaldispense",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.site",
            ),
        ),
        migrations.AddField(
            model_name="historicaldispenseitem",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.site",
            ),
        ),
        migrations.AddField(
            model_name="historicalstock",
            name="allocated",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="historicalstock",
            name="transferred",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="historicalstocktransferconfirmation",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.site",
            ),
        ),
        migrations.AddField(
            model_name="historicalstocktransferconfirmationitem",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.site",
            ),
        ),
        migrations.AddField(
            model_name="historicalstoragebin",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.site",
            ),
        ),
        migrations.AddField(
            model_name="historicalstoragebinitem",
            name="site",
            field=models.ForeignKey(
                blank=True,
                db_constraint=False,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="+",
                to="sites.site",
            ),
        ),
        migrations.AddField(
            model_name="stock",
            name="allocated",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="stock",
            name="transferred",
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name="stocktransferconfirmation",
            name="site",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="sites.site",
            ),
        ),
        migrations.AddField(
            model_name="stocktransferconfirmationitem",
            name="site",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="sites.site",
            ),
        ),
        migrations.AddField(
            model_name="storagebin",
            name="site",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="sites.site",
            ),
        ),
        migrations.AddField(
            model_name="storagebinitem",
            name="site",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="sites.site",
            ),
        ),
    ]



from ..database import Database


class TableToDataframe:
    db_cls = Database

    def __init__(
        self,
        table_name=None,
        db_name=None,
        columns=None,
        limit=None,
        df_handler_cls=None,
        **kwargs,
    ):
        self.db = self.db_cls(db_name=db_name, **kwargs)
        self.table_name = table_name
        df = self.db.select_table(table_name=self.table_name, limit=limit)
        if df_handler_cls:
            df_handler = self.df_handler_cls(
                dataframe=df, db=self.db, table_name=table_name, **kwargs
            )
            df = df_handler.dataframe
        if columns:
            self.dataframe = df.loc[:, columns]
        else:
            self.dataframe = df



# Generated by Django 5.2.3 on 2025-07-09 20:40

from django.db import migrations

import edc_pharmacy.models.stock.confirmation


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0081_historicalconfirmation_confirmation"),
    ]

    operations = [
        migrations.AlterModelManagers(
            name="confirmation",
            managers=[
                ("objects", edc_pharmacy.models.stock.confirmation.Manager()),
            ],
        ),
        migrations.RemoveField(
            model_name="confirmation",
            name="site",
        ),
        migrations.RemoveField(
            model_name="historicalconfirmation",
            name="site",
        ),
    ]



from django.conf import settings

if (
    "django_celery_beat" in settings.INSTALLED_APPS
    or "django_celery_beat.apps.AppConfig" in settings.INSTALLED_APPS
):
    celery_manager = [
        "django_celery_beat.add_clockedschedule",
        "django_celery_beat.add_crontabschedule",
        "django_celery_beat.add_intervalschedule",
        "django_celery_beat.add_periodictask",
        "django_celery_beat.add_periodictasks",
        "django_celery_beat.add_solarschedule",
        "django_celery_beat.change_clockedschedule",
        "django_celery_beat.change_crontabschedule",
        "django_celery_beat.change_intervalschedule",
        "django_celery_beat.change_periodictask",
        "django_celery_beat.change_periodictasks",
        "django_celery_beat.change_solarschedule",
        "django_celery_beat.delete_clockedschedule",
        "django_celery_beat.delete_crontabschedule",
        "django_celery_beat.delete_intervalschedule",
        "django_celery_beat.delete_periodictask",
        "django_celery_beat.delete_periodictasks",
        "django_celery_beat.delete_solarschedule",
        "django_celery_beat.view_clockedschedule",
        "django_celery_beat.view_crontabschedule",
        "django_celery_beat.view_intervalschedule",
        "django_celery_beat.view_periodictask",
        "django_celery_beat.view_periodictasks",
        "django_celery_beat.view_solarschedule",
        "django_celery_results.add_taskresult",
        "django_celery_results.change_taskresult",
        "django_celery_results.delete_taskresult",
        "django_celery_results.view_taskresult",
    ]
else:
    celery_manager = []

if (
    "django_celery_results" in settings.INSTALLED_APPS
    or "django_celery_results.apps.AppConfig" in settings.INSTALLED_APPS
):
    celery_manager.extend(
        [
            "django_celery_results.add_taskresult",
            "django_celery_results.change_taskresult",
            "django_celery_results.delete_taskresult",
            "django_celery_results.view_taskresult",
        ]
    )



from __future__ import annotations

import contextlib
from decimal import Decimal

from django.core.exceptions import ObjectDoesNotExist
from django.db import models, transaction

from edc_lab_results.model_mixin_factories import reportable_result_model_mixin_factory
from edc_registration.models import RegisteredSubject
from edc_reportable.units import EGFR_UNITS, PERCENT
from edc_reportable.utils import get_reference_range_collection

from ..calculators import EgfrCalculatorError
from ..egfr import Egfr


class EgfrModelMixin(
    reportable_result_model_mixin_factory(
        utest_id="egfr",
        verbose_name="eGFR",
        decimal_places=4,
        default_units=EGFR_UNITS,
        max_digits=8,
        units_choices=((EGFR_UNITS, EGFR_UNITS),),
    ),
    reportable_result_model_mixin_factory(
        utest_id="egfr_drop",
        verbose_name="eGFR Drop",
        decimal_places=4,
        default_units=PERCENT,
        max_digits=10,
        units_choices=((PERCENT, PERCENT),),
    ),
    models.Model,
):
    """Declared with a bloodresult RFT CRF model.

    As a lab result CRF, expects subject_visit, requisition
    and report_datetime.

    See edc_lab_result, edc_crf.
    """

    percent_drop_threshold: float = 20.0
    baseline_timepoint: int = 0
    egfr_formula_name: str = None
    egfr_cls = Egfr

    def save(self, *args, **kwargs):
        if self.creatinine_value:
            self.set_egfr_value_or_raise()
        super().save(*args, **kwargs)

    def set_egfr_value_or_raise(self) -> None:
        egfr = self.egfr_cls(**self.egfr_options)
        try:
            self.egfr_value = egfr.egfr_value
        except EgfrCalculatorError:
            if self.creatinine_value:
                raise
            self.egfr_value = None
            self.egfr_units = None
            self.egfr_grade = None
            self.egfr_drop_value = None
            self.egfr_drop_units = None
            self.egfr_drop_grade = None
        else:
            self.egfr_units = egfr.egfr_units
            self.egfr_grade = egfr.egfr_grade
            self.egfr_drop_value = egfr.egfr_drop_value
            self.egfr_drop_units = egfr.egfr_drop_units
            self.egfr_drop_grade = egfr.egfr_drop_grade

    @property
    def egfr_options(self) -> dict:
        rs = RegisteredSubject.objects.get(
            subject_identifier=self.related_visit.subject_identifier
        )
        return dict(
            calling_crf=self,
            dob=rs.dob,
            gender=rs.gender,
            ethnicity=rs.ethnicity,
            weight_in_kgs=self.get_weight_in_kgs_for_egfr(),
            percent_drop_threshold=self.percent_drop_threshold,
            value_threshold=45.0000,
            report_datetime=self.report_datetime,
            baseline_egfr_value=self.get_baseline_egfr_value(),
            formula_name=self.egfr_formula_name,
            reference_range_collection=get_reference_range_collection(self),
        )

    def get_baseline_egfr_value(self) -> float | None:
        """Returns a baseline or reference eGFR value.

        Expects a longitudinal / CRF model with attrs `subject_visit`.
        """
        egfr_value = None
        with transaction.atomic():
            baseline_visit = self.related_visit.__class__.objects.get(
                appointment__subject_identifier=self.related_visit.subject_identifier,
                appointment__visit_schedule_name=self.related_visit.visit_schedule_name,
                appointment__schedule_name=self.related_visit.schedule_name,
                appointment__timepoint=self.baseline_timepoint,
                visit_code_sequence=0,
            )
        with transaction.atomic(), contextlib.suppress(ObjectDoesNotExist):
            egfr_value = self.__class__.objects.get(subject_visit=baseline_visit).egfr_value
        return egfr_value

    def get_weight_in_kgs_for_egfr(self) -> Decimal | None:
        return None

    class Meta:
        abstract = True



# Generated by Django 5.1.2 on 2024-11-19 11:42

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("edc_pharmacy", "0045_historicallot_manufacture_date_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="historicalrx",
            name="slug",
        ),
        migrations.RemoveField(
            model_name="rx",
            name="slug",
        ),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

from django.db import models

from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from ...constants import CRF, REQUIRED
from ...metadata_updater import MetadataUpdater

if TYPE_CHECKING:
    from edc_crf.model_mixins import CrfModelMixin as Base
    from edc_visit_schedule.visit import Visit

    from ...models import CrfMetadata, RequisitionMetadata
    from ..creates import CreatesMetadataModelMixin
    from .updates_crf_metadata_model_mixin import UpdatesCrfMetadataModelMixin
    from .updates_requisition_metadata_model_mixin import (
        UpdatesRequisitionMetadataModelMixin,
    )

    class RelatedVisitModel(CreatesMetadataModelMixin, Base):
        pass

    class CrfModel(UpdatesCrfMetadataModelMixin, Base):
        related_visit = models.ForeignKey(RelatedVisitModel, on_delete=models.PROTECT)

    class RequisitionModel(UpdatesRequisitionMetadataModelMixin, Base):
        related_visit = models.ForeignKey(RelatedVisitModel, on_delete=models.PROTECT)


class MetadataError(Exception):
    pass


class UpdatesMetadataModelMixin(models.Model):
    """A model mixin used on CRF models to enable them to
    update metadata upon save and delete.
    """

    metadata_updater_cls = MetadataUpdater
    metadata_category: str = CRF

    def metadata_update(self: CrfModel | RequisitionModel, entry_status: str = None) -> None:
        """Updates metatadata."""
        self.metadata_updater.get_and_update(entry_status=entry_status)

    def run_metadata_rules_for_related_visit(
        self: CrfModel | RequisitionModel, allow_create: bool | None = None
    ) -> None:
        """Runs all the metadata rules for this timepoint."""
        self.related_visit.run_metadata_rules(allow_create=allow_create)

    @property
    def metadata_updater(self: CrfModel | RequisitionModel) -> MetadataUpdater:
        """Returns an instance of MetadataUpdater.

        Override
        """
        raise NotImplementedError("Method not implemented")

    @property
    def metadata_query_options(self: CrfModel | RequisitionModel) -> dict:
        options = self.related_visit.metadata_query_options
        options.update(
            {
                "subject_identifier": self.related_visit.subject_identifier,
                "model": self._meta.label_lower,
            }
        )
        return options

    @property
    def metadata_model(
        self: CrfModel | RequisitionModel,
    ) -> type[CrfMetadata] | type[RequisitionMetadata]:
        """Returns the metadata model associated with self.

        Override
        """
        raise NotImplementedError("Method not implemented")

    @property
    def metadata_default_entry_status(self: CrfModel | RequisitionModel) -> str:
        """Returns a string that represents the default entry status
        of the CRF in the visit schedule.

        Override
        """
        raise NotImplementedError("Method not implemented")

    def metadata_reset_on_delete(self: CrfModel | RequisitionModel) -> None:
        """Sets this model instance`s metadata model instance
        to its original entry_status.
        """
        obj = self.metadata_model.objects.get(**self.metadata_query_options)
        try:
            obj.entry_status = self.metadata_default_entry_status
        except IndexError:
            # if IndexError, implies CRF is not listed in
            # the visit schedule, so remove it.
            # for example, this is a PRN form
            obj.delete()
        else:
            obj.entry_status = self.metadata_default_entry_status or REQUIRED
            obj.report_datetime = None
            obj.save()

    @property
    def metadata_visit_object(self: CrfModel | RequisitionModel) -> Visit:
        visit_schedule = site_visit_schedules.get_visit_schedule(
            visit_schedule_name=self.related_visit.visit_schedule_name
        )
        schedule = visit_schedule.schedules.get(self.related_visit.schedule_name)
        return schedule.visits.get(self.related_visit.visit_code)

    class Meta:
        abstract = True



from django.core.management.base import BaseCommand, CommandError

from ...import_holidays import HolidayImportError, import_holidays


class Command(BaseCommand):
    help = "Import country holidays"

    def handle(self, *args, **options):
        try:
            import_holidays(verbose=True)
        except HolidayImportError as e:
            raise CommandError(e)



from django.db import models

from edc_metadata.metadata_helper import MetadataHelperMixin


class MetadataHelperModelMixin(MetadataHelperMixin, models.Model):
    """The functionality here is mostly accessed in a signal.

    See edc_metadata/models/signals"""

    class Meta:
        abstract = True
